import json
import torch

from torch import Tensor
from typing import Dict
from typing import List
from typing import Union
from typing import Optional
from cftool.misc import shallow_copy_dict
from cftool.array import tensor_dict_type

from ..misc.toolkit import get_tensors
from ..misc.toolkit import download_static
from ..api.cv.diffusion import DiffusionAPI


key_mapping_inv = {}


def _convert_cond_stage(d: tensor_dict_type, md: tensor_dict_type) -> tensor_dict_type:
    def _get_dv(k: str) -> Tensor:
        v = d.get(k)
        if v is None:
            v = d[k.replace(".text_model", "")]
        return v

    if not d or not md:
        return {}
    d, md = map(shallow_copy_dict, [d, md])
    nd = {
        "condition_model.m.token_embedding.weight": _get_dv(
            "cond_stage_model.transformer.text_model.embeddings.token_embedding.weight"
        ),
        "condition_model.m.text_transformer.encoder.pos_encoding.pos_encoding": _get_dv(
            "cond_stage_model.transformer.text_model.embeddings.position_embedding.weight"
        )[None, ...],
        "condition_model.m.text_transformer.encoder.head.norms.0.weight": _get_dv(
            "cond_stage_model.transformer.text_model.final_layer_norm.weight"
        ),
        "condition_model.m.text_transformer.encoder.head.norms.0.bias": _get_dv(
            "cond_stage_model.transformer.text_model.final_layer_norm.bias"
        ),
    }
    for key in [
        "condition_model.m.logit_scale",
        "condition_model.m.text_transformer.attention_mask",
        "condition_model.m.text_projection.weight",
        "condition_model.m.text_projection.bias",
    ]:
        nd[key] = md.pop(key)
    prefix = "cond_stage_model.transformer.text_model.encoder.layers"
    m_prefix = "condition_model.m.text_transformer.encoder.mixing_blocks"
    layer_indices = []
    for key in d:
        if not key.startswith(prefix):
            continue
        try:
            layer_indices.append(int(key.split(".")[5]))
        except:
            continue
    num_cond_layers = max(layer_indices) + 1
    for i in range(num_cond_layers):
        i_prefix = f"{prefix}.{i}"
        m_i_prefix = f"{m_prefix}.{i}"
        get_w = lambda n: _get_dv(f"{i_prefix}.self_attn.{n}_proj.weight")
        get_b = lambda n: _get_dv(f"{i_prefix}.self_attn.{n}_proj.bias")
        in_w = torch.cat(list(map(get_w, ["q", "k", "v"])))
        qkv_bias = torch.cat(list(map(get_b, ["q", "k", "v"])))
        m_i_token_mixing_prefix = f"{m_i_prefix}.token_mixing.net"
        nd[f"{m_i_token_mixing_prefix}.in_w"] = in_w
        nd[f"{m_i_token_mixing_prefix}.qkv_bias"] = qkv_bias
        out_w = _get_dv(f"{i_prefix}.self_attn.out_proj.weight")
        out_b = _get_dv(f"{i_prefix}.self_attn.out_proj.bias")
        nd[f"{m_i_token_mixing_prefix}.out_linear.linear.weight"] = out_w
        nd[f"{m_i_token_mixing_prefix}.out_linear.linear.bias"] = out_b
        get_fc_w = lambda n: _get_dv(f"{i_prefix}.mlp.{n}.weight")
        get_fc_b = lambda n: _get_dv(f"{i_prefix}.mlp.{n}.bias")
        m_i_channel_mixing_prefix = f"{m_i_prefix}.channel_mixing.net"
        nd[f"{m_i_channel_mixing_prefix}.0.linear.weight"] = get_fc_w("fc1")
        nd[f"{m_i_channel_mixing_prefix}.0.linear.bias"] = get_fc_b("fc1")
        nd[f"{m_i_channel_mixing_prefix}.3.linear.weight"] = get_fc_w("fc2")
        nd[f"{m_i_channel_mixing_prefix}.3.linear.bias"] = get_fc_b("fc2")
        for ln, m_ln in zip(
            ["layer_norm1", "layer_norm2"],
            ["token_norm", "channel_norm"],
        ):
            nd[f"{m_i_prefix}.{m_ln}.weight"] = _get_dv(f"{i_prefix}.{ln}.weight")
            nd[f"{m_i_prefix}.{m_ln}.bias"] = _get_dv(f"{i_prefix}.{ln}.bias")
    return nd


def _convert_first_stage(d: tensor_dict_type, md: tensor_dict_type) -> tensor_dict_type:
    d, md = map(shallow_copy_dict, [d, md])
    keys = list(d.keys())
    m_keys = list(md.keys())
    for k in [
        "first_stage.codebook.embedding.weight",
        "first_stage.to_embedding.weight",
        "first_stage.to_embedding.bias",
        "first_stage.from_embedding.weight",
        "first_stage.from_embedding.bias",
    ]:
        if k in m_keys:
            m_keys.remove(k)
            m_keys.append(k)
    start_idx = None
    end_idx = None
    for i, k in enumerate(keys):
        if start_idx is None and k.startswith("first_stage_model.decoder.up"):
            start_idx = i
        elif k == "first_stage_model.decoder.norm_out.weight":
            end_idx = i
    if end_idx is None:
        raise ValueError("`first_stage_model.decoder.norm_out.weight` is missing")
    before = keys[:start_idx]
    up = keys[start_idx:end_idx]
    after = keys[end_idx:]
    num_up = int(keys[end_idx - 1].split(".")[3]) + 1
    new_up = []
    mapping = shallow_copy_dict(d)
    local_key_mapping = {k: k for k in d}
    for k in up:
        mapping.pop(k)
    for i in reversed(range(num_up)):
        for k in up:
            if k.startswith(f"first_stage_model.decoder.up.{i}"):
                ks = k.split(".")
                ks[3] = str(num_up - int(ks[3]))
                new_k = ".".join(ks)
                new_up.append(new_k)
                mapping[new_k] = d[k]
                local_key_mapping[new_k] = k
    keys = before + new_up + after

    def _handle_attn(prefix: str) -> None:
        attn_keys: Dict[int, Dict[int, List[str]]] = {}
        all_attn_keys = []
        for key in keys:
            if key.startswith(prefix):
                ks = key.split(".")
                if ks[4] == "attn":
                    level = int(ks[3])
                    idx = int(ks[5])
                    attn_keys.setdefault(level, {}).setdefault(idx, []).append(key)
                    all_attn_keys.append(key)
        for level, l_attn_keys in attn_keys.items():
            for idx, idx_attn_keys in l_attn_keys.items():
                for i, key in enumerate(keys):
                    if key.startswith(f"{prefix}.{level}.block.{idx + 1}"):
                        for attn_key in idx_attn_keys:
                            keys.remove(attn_key)
                            keys.insert(i, attn_key)
                            i += 1
                        break

    _handle_attn("first_stage_model.encoder.down")
    _handle_attn("first_stage_model.decoder.up")
    new_d = {}
    for k, mk in zip(keys, m_keys):
        v = mapping[k]
        mv = md[mk]
        assert v.shape == mv.shape, f"{k} ({v.shape}) != {mk} ({mv.shape})"
        new_d[mk] = v
        key_mapping_inv[mk] = local_key_mapping[k]
    return new_d


def _convert_others(d: tensor_dict_type, md: tensor_dict_type) -> tensor_dict_type:
    nd = {}
    d, md = map(shallow_copy_dict, [d, md])
    keys = list(d.keys())
    m_keys = list(md.keys())
    for k in reversed(keys):
        if k == "model_ema.decay":
            keys.remove(k)
        elif not k.startswith("model") and not k.startswith("model_ema"):
            keys.remove(k)
    for mk in reversed(m_keys):
        if not mk.startswith("unet") and not mk.startswith("unet_ema"):
            continue
    keys.remove("model_ema.num_updates")
    keys.append("model_ema.num_updates")
    for k, mk in zip(keys, m_keys):
        v = d[k]
        mv = md[mk]
        assert v.shape == mv.shape, f"{k} ({v.shape}) != {mk} ({mv.shape})"
        nd[mk] = v
        key_mapping_inv[mk] = k
    return nd


def _convert(d: tensor_dict_type, md: tensor_dict_type) -> tensor_dict_type:
    key_mapping_inv.clear()
    d, md = map(shallow_copy_dict, [d, md])
    nd = {}
    d = {k: v.cpu() for k, v in d.items()}
    md = {k: v.cpu() for k, v in md.items()}
    if "cond_stage_model.channel_mapper.weight" in d:
        print(">  [info] injecting channel_mapper")
        assert (
            d["cond_stage_model.channel_mapper.weight"].shape
            == md["condition_model.channel_mapper.weight"].shape
        )
        k = "cond_stage_model.channel_mapper.weight"
        mk = "condition_model.channel_mapper.weight"
        nd[mk] = d[k]
        key_mapping_inv[mk] = k
    # condition
    cond_d = {}
    for k in list(d.keys()):
        if k.startswith("cond_stage_model"):
            cond_d[k] = d.pop(k)
    cond_md = {k: v for k, v in md.items() if k.startswith("condition_model")}
    for k in cond_md:
        md.pop(k)
    nd.update(_convert_cond_stage(cond_d, cond_md))
    # first stage
    fd = {k: v for k, v in d.items() if k.startswith("first_stage_model")}
    fmd = {k: v for k, v in md.items() if k.startswith("first_stage")}
    nd.update(_convert_first_stage(fd, fmd))
    # others
    od = {k: v for k, v in d.items() if not k.startswith("first_stage_model")}
    omd = {k: v for k, v in md.items() if not k.startswith("first_stage")}
    nd.update(_convert_others(od, omd))
    # ema
    ema_num_updates_k = "unet_ema.num_updates"
    if ema_num_updates_k in nd:
        nd.pop(ema_num_updates_k)
        key_mapping_inv.pop(ema_num_updates_k)
        normal_keys = [k for k in list(nd) if k.startswith("unet.")]
        ema_keys = [k for k in list(nd) if k.startswith("unet_ema")]
        for k, ema_k in zip(normal_keys, ema_keys):
            assert nd.pop(k).shape == nd[ema_k].shape
        for k, ema_k in zip(normal_keys, ema_keys):
            nd[k] = nd.pop(ema_k)
            key_mapping_inv[k] = key_mapping_inv.pop(ema_k)
    return nd


def _convert_with_key_mapping(
    d: tensor_dict_type,
    md: tensor_dict_type,
) -> tensor_dict_type:
    d, md = map(shallow_copy_dict, [d, md])
    with open(download_static("sd_mapping", extension="json"), "r") as f:
        mapping = json.load(f)
    nd = _convert_cond_stage(d, md)
    for k, mk in mapping.items():
        nd[mk] = d[k]
    for k, v in md.items():
        nd.setdefault(k, v)
    return nd


def _get_inp(
    inp: Union[str, tensor_dict_type],
    vae_inp: Optional[Union[str, tensor_dict_type]],
) -> tensor_dict_type:
    inp = get_tensors(inp)
    if vae_inp is not None:
        vae_inp = get_tensors(vae_inp)
        num_injected = 0
        for k, v in vae_inp.items():
            inp_k = f"first_stage_model.{k}"
            if inp_k in inp:
                inp[inp_k] = v
                num_injected += 1
        print(f">> injected vae parameters: {num_injected}")
    return inp


def convert(
    inp: Union[str, tensor_dict_type],
    api: DiffusionAPI,
    *,
    vae_inp: Optional[Union[str, tensor_dict_type]] = None,
    load: bool = False,
    use_mapping: bool = True,
) -> tensor_dict_type:
    inp = _get_inp(inp, vae_inp)
    with api.load_context() as m:
        md = m.state_dict()
        if not use_mapping:
            nd = _convert(inp, md)
        else:
            nd = _convert_with_key_mapping(inp, md)
        if load:
            m.load_state_dict(nd)
    return nd


def convert_v2(
    inp: Union[str, tensor_dict_type],
    api: DiffusionAPI,
    *,
    vae_inp: Optional[Union[str, tensor_dict_type]] = None,
    load: bool = False,
) -> tensor_dict_type:
    inp = _get_inp(inp, vae_inp)
    for k in [
        "log_one_minus_alphas_cumprod",
        "sqrt_recip_alphas_cumprod",
        "sqrt_recipm1_alphas_cumprod",
    ]:
        inp.pop(k)
    with api.load_context() as m:
        md = m.state_dict()
        nd = {}
        for k in [
            "condition_model.m.text_projection.bias",
            "condition_model.m.text_transformer.attention_mask",
        ]:
            nd[k] = md.pop(k)
        with open(download_static("sd_v2_mapping", extension="json"), "r") as f:
            mapping = json.load(f)
        for k, mk in mapping.items():
            v = inp[k]
            if mk == "condition_model.m.text_projection.weight":
                v = v.t()
            nd[mk] = v.view(md[mk].shape)
        if load:
            m.load_state_dict(nd)
    return nd


def inject(inp: Union[str, tensor_dict_type], api: DiffusionAPI) -> None:
    d = get_tensors(inp)
    with api.load_context() as m:
        m.load_state_dict(d)


def convert_controlnet(inp: Union[str, tensor_dict_type]) -> tensor_dict_type:
    inp = get_tensors(inp)
    with open(download_static("sd_controlnet_mapping", extension="json"), "r") as f:
        mapping = json.load(f)
    nd = {}
    for k, v in inp.items():
        if "control_model" in k:
            nd[mapping[k]] = v
    return nd


__all__ = [
    "convert",
    "convert_v2",
    "inject",
    "convert_controlnet",
]



import torch.nn as nn

from abc import abstractmethod
from abc import ABCMeta
from torch import Tensor
from typing import Any
from typing import Dict
from typing import Type
from cftool.misc import WithRegister

from .....schema import IDLModel


# `condition_models` are rather 'independent' models, they can be used independently
condition_models: Dict[str, Type["IConditionModel"]] = {}
# `specialized_condition_models` are specialized for conditioning, they cannot be used independently
specialized_condition_models: Dict[str, Type[nn.Module]] = {}


class IConditionModel(IDLModel, WithRegister, metaclass=ABCMeta):
    d = condition_models

    def __init__(self, m: nn.Module):
        super().__init__()
        self.m = m

    @abstractmethod
    def forward(self, cond: Any) -> Tensor:
        pass


class ISpecializedConditionModel(nn.Module, WithRegister, metaclass=ABCMeta):
    d = specialized_condition_models

    @abstractmethod
    def forward(self, cond: Any) -> Tensor:
        pass


__all__ = [
    "condition_models",
    "specialized_condition_models",
    "IConditionModel",
    "ISpecializedConditionModel",
]



import os
import sys

from typing import Any
from typing import Dict
from typing import Type
from typing import Union
from typing import Optional
from cftool.misc import Serializer
from cftool.misc import ISerializable
from cftool.misc import PureFromInfoMixin

from ...schema import DLConfig


class Task(PureFromInfoMixin, ISerializable):
    d: Dict[str, Type["Task"]] = {}
    config: Optional[Dict[str, Any]]
    run_command: Optional[str]
    meta_kwargs: Dict[str, Any]

    @classmethod
    def init(
        cls,
        config: Optional[DLConfig],
        run_command: Optional[str] = None,
        **meta_kwargs: Any,
    ) -> "Task":
        self = cls()
        if config is None:
            self.config = None
        else:
            self.config = config.to_pack().asdict()
        self.run_command = run_command
        self.meta_kwargs = meta_kwargs
        return self

    def run(
        self,
        execute: str,
        task_folder: str,
        cuda: Optional[Union[int, str]],
    ) -> "Task":
        if self.run_command is not None:
            command = self.run_command
        else:
            command = f"{sys.executable} -m cflearn.dist.ml.runs.{execute}"
        self.meta_kwargs["cuda"] = cuda
        os.makedirs(task_folder, exist_ok=True)
        Serializer.save(task_folder, self)
        os.system(f"{command} --task_folder {task_folder}")
        return self

    def to_info(self) -> Dict[str, Any]:
        return dict(
            config=self.config,
            run_command=self.run_command,
            meta_kwargs=self.meta_kwargs,
        )


Task.register("task")(Task)


__all__ = ["Task"]



import numpy as np

from typing import Tuple

from ..blocks import ColumnTypes

try:
    from sklearn.utils import Bunch
    from sklearn.datasets import load_iris
    from sklearn.datasets import load_digits
    from sklearn.datasets import load_breast_cancer
    from sklearn.datasets import fetch_california_housing
except:
    Bunch = None
    load_iris = load_digits = load_breast_cancer = fetch_california_housing = None


TXYDataset = Tuple[np.ndarray, np.ndarray]


def _from_bunch(bunch: Bunch, ctype: ColumnTypes) -> TXYDataset:
    x = bunch.data.astype(np.float64)
    y_dtype = np.float64 if ctype == ColumnTypes.NUMERICAL else int
    y = bunch.target.reshape([-1, 1]).astype(y_dtype)
    return x, y


def iris_dataset() -> TXYDataset:
    return _from_bunch(load_iris(), ColumnTypes.CATEGORICAL)


def digits_dataset() -> TXYDataset:
    return _from_bunch(load_digits(), ColumnTypes.CATEGORICAL)


def california_dataset() -> TXYDataset:
    return _from_bunch(fetch_california_housing(), ColumnTypes.NUMERICAL)


def breast_dataset() -> TXYDataset:
    return _from_bunch(load_breast_cancer(), ColumnTypes.CATEGORICAL)


__all__ = [
    "iris_dataset",
    "digits_dataset",
    "california_dataset",
    "breast_dataset",
]



import math
import torch

import numpy as np
import torch.nn as nn
import torch.nn.functional as F

from torch import Tensor
from typing import Any
from typing import Dict
from typing import List
from typing import Union
from typing import Optional
from typing import NamedTuple
from cftool.array import to_torch
from cftool.types import arr_type

from ...schema import MLEncoderSettings
from ...schema import MLGlobalEncoderSettings
from ...misc.toolkit import shallow_copy_dict
from ...misc.toolkit import Initializer
from ...modules.blocks import Lambda


class OneHot(Lambda):
    def __init__(self, dim: int):
        self.dim = dim
        one_hot_fn = lambda column: F.one_hot(column, dim)
        super().__init__(one_hot_fn, f"one_hot_{dim}")

    def forward(self, net: Tensor) -> Tensor:
        return super().forward(net).to(torch.float32)


class EmbeddingConfig(NamedTuple):
    out_dim: int
    init_method: Optional[str]
    init_config: Dict[str, Any]
    dropout: float


class Embedding(nn.Module):
    def __init__(self, in_dim: int, config: EmbeddingConfig):
        super().__init__()
        weights = torch.empty(in_dim, config.out_dim)
        if config.init_method is None:
            nn.init.normal_(weights)
        else:
            Initializer(config.init_config).initialize(weights, config.init_method)
        self.weights = nn.Parameter(weights)
        if config.dropout <= 0.0 or 1.0 <= config.dropout:
            self.dropout = None
        else:
            self.dropout = nn.Dropout(config.dropout, inplace=True)
        self.in_dim = in_dim
        self.out_dim = config.out_dim

    def extra_repr(self) -> str:
        return f"{self.in_dim} -> {self.out_dim}"

    def forward(self, net: Tensor) -> Tensor:
        net = F.embedding(net, self.weights)
        if self.dropout is not None:
            net = self.dropout(net)
        return net


class EncodingResult(NamedTuple):
    indices: Tensor
    one_hot: Optional[Tensor]
    embedding: Optional[Tensor]

    @property
    def merged(self) -> Optional[Tensor]:
        if self.one_hot is None and self.embedding is None:
            return None
        if self.one_hot is None:
            assert self.embedding is not None
            return self.embedding
        if self.embedding is None:
            assert self.one_hot is not None
            return self.one_hot
        return torch.cat([self.one_hot, self.embedding], dim=-1)


def get_embedding_config(
    in_dim: int,
    kw: Dict[str, Any],
    global_encoder_settings: Optional[MLGlobalEncoderSettings],
) -> EmbeddingConfig:
    ges = global_encoder_settings or MLGlobalEncoderSettings()
    default_out_dim: Union[str, int]
    if ges.embedding_dim is None:
        default_out_dim = "auto"
    else:
        default_out_dim = ges.embedding_dim
    # get out_dim
    out_dim = kw.get("out_dim", default_out_dim)
    if isinstance(out_dim, int):
        out_dim = out_dim
    elif out_dim == "log":
        out_dim = math.ceil(math.log2(in_dim))
    elif out_dim == "sqrt":
        out_dim = math.ceil(math.sqrt(in_dim))
    elif out_dim == "auto":
        out_dim = max(4, min(8, math.ceil(math.log2(in_dim))))
    else:
        raise ValueError(f"embedding dim '{out_dim}' is not defined")
    # get init arguments
    init_method = kw.get("init_method", "truncated_normal")
    init_config = kw.get("init_config", {"mean": 0.0, "std": 0.02})
    # get dropout
    dropout = kw.get(
        "dropout",
        0.1 if ges.embedding_dropout is None else ges.embedding_dropout,
    )
    # return
    return EmbeddingConfig(out_dim, init_method, init_config, dropout)


def to_split(columns: Tensor) -> List[Tensor]:
    return list(columns.unbind(dim=-1))


class Encoder(nn.Module):
    dims: Tensor

    def __init__(
        self,
        settings: Dict[str, MLEncoderSettings],
        global_encoder_settings: Optional[MLGlobalEncoderSettings] = None,
    ):
        super().__init__()
        self.embeddings = nn.ModuleDict()
        self.one_hot_encoders = nn.ModuleDict()
        dims = []
        self.tgt_columns = []
        self.one_hot_columns = []
        self.embedding_columns = []
        self.one_hot_dim = 0
        self.embedding_dim = 0
        self.dim_increment = 0
        for str_idx in sorted(settings):
            idx = int(str_idx)
            setting = settings[str_idx]
            dim = setting.dim
            dims.append(dim)
            self.tgt_columns.append(idx)
            if setting.use_one_hot:
                self.one_hot_columns.append(idx)
                self.one_hot_encoders[str_idx] = OneHot(dim)
                self.one_hot_dim += dim
                self.dim_increment += dim - 1
            if setting.use_embedding:
                self.embedding_columns.append(idx)
                kw = shallow_copy_dict(setting.method_configs or {})
                config = get_embedding_config(dim, kw, global_encoder_settings)
                self.embeddings[str_idx] = Embedding(dim, config)
                self.embedding_dim += config.out_dim
                self.dim_increment += config.out_dim - 1
        # setup useful properties
        self.categorical_dim = self.one_hot_dim + self.embedding_dim
        self.num_one_hot = len(self.one_hot_encoders)
        self.num_embedding = len(self.embeddings)
        self.use_one_hot = self.num_one_hot > 0
        self.use_embedding = self.num_embedding > 0
        self.is_empty = not self.use_one_hot and not self.use_embedding
        self.all_one_hot = len(self.one_hot_columns) == len(self.tgt_columns)
        self.all_embedding = len(self.embedding_columns) == len(self.tgt_columns)
        self.register_buffer("dims", torch.tensor(dims, dtype=torch.float32))

    # api

    # encode all columns
    def forward(self, x_batch: Tensor) -> EncodingResult:
        # extract categorical
        categorical_columns = x_batch[..., self.tgt_columns]
        # oob imputation
        oob_mask = categorical_columns >= self.dims
        if oob_mask.any().item():
            categorical_columns = torch.where(
                oob_mask,
                torch.zeros_like(categorical_columns),
                categorical_columns,
            )
        # turn to long indices
        indices = categorical_columns.to(torch.long)
        # one hot
        if not self.use_one_hot:
            one_hot = None
        else:
            if self.all_one_hot:
                one_hot_indices = indices
            else:
                one_hot_indices = indices[..., self.one_hot_columns]
            one_hot = self._one_hot(one_hot_indices)
        # embedding
        if not self.use_embedding:
            embedding = None
        else:
            if self.all_embedding:
                embedding_indices = indices
            else:
                embedding_indices = indices[..., self.embedding_columns]
            embedding = self._embedding(embedding_indices)
        # return
        return EncodingResult(indices, one_hot, embedding)

    # encode single column
    def encode(self, column: arr_type, column_idx: int) -> EncodingResult:
        if column_idx not in self.tgt_columns:
            return EncodingResult(None, None, None)
        if isinstance(column, np.ndarray):
            column = to_torch(column)
        indices = column.to(torch.long)
        flat_indices = indices.view(-1)
        # one hot
        if not self.use_one_hot or column_idx not in self.one_hot_columns:
            one_hot = None
        else:
            one_hot = self.one_hot_encoders[str(column_idx)](flat_indices)
        # embedding
        if not self.use_embedding or column_idx not in self.embedding_columns:
            embedding = None
        else:
            embedding_indices = flat_indices.clone()
            embedding = self.embeddings[str(column_idx)](embedding_indices)
            if self.embedding_dropout is not None:
                embedding = self.embedding_dropout(embedding)
        return EncodingResult(indices, one_hot, embedding)

    # internal

    def _one_hot(self, one_hot_indices: Tensor) -> Tensor:
        split = to_split(one_hot_indices)
        encodings = [
            self.one_hot_encoders[str(self.one_hot_columns[i])](flat_feature)
            for i, flat_feature in enumerate(split)
        ]
        return torch.cat(encodings, dim=-1)

    def _embedding(self, embedding_indices: Tensor) -> Tensor:
        split = to_split(embedding_indices)
        encodings = [
            self.embeddings[str(self.embedding_columns[i])](flat_feature)
            for i, flat_feature in enumerate(split)
        ]
        return torch.cat(encodings, dim=-1)


__all__ = [
    "Encoder",
    "EncodingResult",
]



import torch

from torch import Tensor
from typing import Dict
from typing import List
from typing import Optional
from cftool.types import tensor_dict_type

from .base import MLModel
from .fcnn import FCNN
from .linear import Linear
from ...schema import MLEncoderSettings
from ...schema import MLGlobalEncoderSettings


@MLModel.register("wnd")
class WideAndDeep(MLModel):
    def __init__(
        self,
        input_dim: int,
        output_dim: int,
        num_history: int = 1,
        hidden_units: Optional[List[int]] = None,
        *,
        mapping_type: str = "basic",
        bias: bool = True,
        activation: str = "ReLU",
        batch_norm: bool = False,
        dropout: float = 0.0,
        encoder_settings: Optional[Dict[str, MLEncoderSettings]] = None,
        global_encoder_settings: Optional[MLGlobalEncoderSettings] = None,
    ):
        super().__init__(
            encoder_settings=encoder_settings,
            global_encoder_settings=global_encoder_settings,
        )
        encoder = self.encoder
        if encoder is None or encoder.is_empty:
            wide_dim = deep_dim = input_dim
        else:
            wide_dim = encoder.categorical_dim
            numerical_dim = input_dim - encoder.num_one_hot - encoder.num_embedding
            deep_dim = numerical_dim + encoder.embedding_dim
        self.fcnn = FCNN(
            deep_dim,
            output_dim,
            num_history,
            hidden_units,
            mapping_type=mapping_type,
            bias=bias,
            activation=activation,
            batch_norm=batch_norm,
            dropout=dropout,
        )
        self.linear = Linear(wide_dim, output_dim, num_history, bias=bias)

    def forward(self, net: Tensor) -> tensor_dict_type:
        if len(net.shape) > 2:
            net = net.contiguous().view(len(net), -1)
        encoded = self.encode(net)
        one_hot = encoded.one_hot
        embedding = encoded.embedding
        numerical = encoded.numerical
        # wide
        if one_hot is None and embedding is None:
            wide_inp = numerical
        else:
            if one_hot is None:
                wide_inp = embedding
            elif embedding is None:
                wide_inp = one_hot
            else:
                wide_inp = torch.cat([one_hot, embedding], dim=-1)
        wide_net = self.linear(wide_inp)
        # deep
        if embedding is None:
            deep_inp = numerical
        elif numerical is None:
            deep_inp = embedding
        else:
            deep_inp = torch.cat([numerical, embedding], dim=-1)
        deep_net = self.fcnn(deep_inp)
        # merge
        return wide_net + deep_net


__all__ = ["WideAndDeep"]



import numpy as np

from typing import Any
from typing import List
from typing import Type
from typing import Union

try:
    from transformers import BertTokenizer
    from transformers import PreTrainedTokenizer
    from transformers import CLIPTokenizer as _CLIPTokenizer
except:
    BertTokenizer = PreTrainedTokenizer = _CLIPTokenizer = None

from .schema import ITokenizer


class ICLIPTokenizer(ITokenizer):
    tag: str
    base: Type[PreTrainedTokenizer]

    def __init__(self, *, pad_to_max: bool = True):
        if self.base is None:
            raise ValueError("`transformers` is needed for `ICLIPTokenizer`")
        self.tokenizer = self.base.from_pretrained(self.tag)
        self.bos_token_id = self.tokenizer.bos_token_id
        self.eos_token_id = self.tokenizer.eos_token_id
        self.comma_token_id = self.tokenizer.get_vocab().get(",</w>")
        self.pad_to_max = pad_to_max

    def encode(
        self,
        texts: List[str],
        *,
        padding: Union[str, bool] = False,
        truncation: bool = True,
        add_special_tokens: bool = True,
        return_numpy: bool = False,
    ) -> Any:
        tokenized = self.tokenizer(
            texts,
            padding=padding,
            truncation=truncation,
            add_special_tokens=add_special_tokens,
        )
        if not return_numpy:
            return tokenized
        return np.array(tokenized["input_ids"], np.int64)

    def tokenize(self, texts: Union[str, List[str]], **kwargs: Any) -> np.ndarray:
        if isinstance(texts, str):
            texts = [texts]
        kwargs["return_numpy"] = True
        kwargs["padding"] = "max_length" if self.pad_to_max else True
        return self.encode(texts, **kwargs)


@ITokenizer.register("clip")
class CLIPTokenizer(ICLIPTokenizer):
    tag = "openai/clip-vit-large-patch14"
    base = _CLIPTokenizer


@ITokenizer.register("clip.chinese")
class ChineseCLIPTokenizer(ICLIPTokenizer):
    tag = "IDEA-CCNL/Taiyi-CLIP-Roberta-large-326M-Chinese"
    base = BertTokenizer

    def __init__(
        self,
        *,
        pad_to_max: bool = False,
    ):
        super().__init__(pad_to_max=pad_to_max)
        self.max_length = 512


__all__ = [
    "ICLIPTokenizer",
    "CLIPTokenizer",
    "ChineseCLIPTokenizer",
]



from .base import BMCONFIGS
from ..model.backboned import HRNetIHModel
from ..model.backboned import DeepLabIHModel


MCONFIGS = {
    "hrnet18s_idih256": {
        "model": HRNetIHModel,
        "params": {"base_config": BMCONFIGS["improved_dih256"]},
    },
    "hrnet18s_v2p_idih256": {
        "model": HRNetIHModel,
        "params": {
            "base_config": BMCONFIGS["improved_dih256"],
            "pyramid_channels": 256,
        },
    },
    "hrnet18_idih256": {
        "model": HRNetIHModel,
        "params": {"base_config": BMCONFIGS["improved_dih256"], "small": False},
    },
    "hrnet18_v2p_idih256": {
        "model": HRNetIHModel,
        "params": {
            "base_config": BMCONFIGS["improved_dih256"],
            "small": False,
            "pyramid_channels": 256,
        },
    },
    "hrnet32_idih256": {
        "model": HRNetIHModel,
        "params": {
            "base_config": BMCONFIGS["improved_dih256"],
            "width": 32,
            "small": False,
        },
    },
    "deeplab_r34_idih256": {
        "model": DeepLabIHModel,
        "params": {"base_config": BMCONFIGS["improved_dih256"]},
    },
    "hrnet18_idih512": {
        "model": HRNetIHModel,
        "params": {
            "base_config": BMCONFIGS["improved_dih512"],
            "small": False,
            "downsize_hrnet_input": True,
        },
    },
    "hrnet18_sedih512": {
        "model": HRNetIHModel,
        "params": {
            "base_config": BMCONFIGS["improved_sedih512"],
            "small": False,
            "downsize_hrnet_input": True,
        },
    },
}



import numpy as np

from abc import ABCMeta
from typing import List
from typing import Type
from typing import Tuple
from typing import Union
from typing import Optional
from dataclasses import dataclass
from cftool.misc import safe_execute
from cftool.types import np_dict_type

from ..blocks import MLBatch
from ..blocks import MLDatasetTag
from ..blocks import GatherBlock
from ..blocks import FileParserBlock
from ..blocks import MLFileProcessorConfig
from ..blocks import MLSplitterConfig
from ..blocks import SplitterBlock
from ..blocks import RecognizerBlock
from ..blocks import MLRecognizerConfig
from ..blocks import NanHandlerBlock
from ..blocks import MLNanHandlerConfig
from ..blocks import PreProcessorBlock
from ..blocks import MLPreProcessConfig
from ..utils import ArrayLoader
from ..utils import IArrayDataset
from ...types import data_type
from ...types import sample_weights_type
from ...schema import IData
from ...schema import IDataBlock
from ...schema import DataConfig
from ...schema import DataProcessor
from ...schema import DataProcessorConfig
from ...constants import INPUT_KEY
from ...constants import LABEL_KEY


@dataclass
@DataProcessorConfig.register("ml.bundled")
class MLBundledProcessorConfig(
    MLSplitterConfig,
    MLPreProcessConfig,
    MLRecognizerConfig,
    MLNanHandlerConfig,
    MLFileProcessorConfig,
    DataProcessorConfig,
):
    """This config is designed to bundle all possible capabilities (e.g. carefree)."""

    @property
    def default_blocks(self) -> List[IDataBlock]:
        return [
            FileParserBlock(),
            NanHandlerBlock(),
            RecognizerBlock(),
            PreProcessorBlock(),
            SplitterBlock(),
        ]


@dataclass
@DataProcessorConfig.register("ml.advanced")
class MLAdvancedProcessorConfig(MLBundledProcessorConfig):
    """
    This config is designed to be capable of any situation,
    but requires users have deeper understanding of the Processor system.
    > e.g. they often need to sepcify the `block_names` explicitly.
    """

    @property
    def default_blocks(self) -> List[IDataBlock]:
        return []


@DataProcessor.register("ml")
class MLDataProcessor(DataProcessor):
    def before_build_in_init(self) -> None:
        self.config.add_blocks(GatherBlock())

    def get_num_samples(self, x: np.ndarray, tag: MLDatasetTag) -> int:
        return len(x)

    def fetch_batch(
        self,
        x: np.ndarray,
        y: Optional[np.ndarray],
        indices: Union[int, List[int], np.ndarray],
        tag: MLDatasetTag,
    ) -> MLBatch:
        return MLBatch(x[indices], None if y is None else y[indices])


class MLDataset(IArrayDataset):
    others: np_dict_type

    def __init__(
        self,
        x: np.ndarray,
        y: Optional[np.ndarray],
        tag: MLDatasetTag,
        processor: MLDataProcessor,
        **others: np.ndarray,
    ):
        super().__init__()
        self.x = x
        self.y = y
        self.tag = tag
        self.processor = processor
        self.others = others

    def __len__(self) -> int:
        return self.processor.get_num_samples(self.x, self.tag)

    def __getitem__(self, item: Union[int, List[int], np.ndarray]) -> np_dict_type:
        kw = dict(x=self.x, y=self.y, indices=item, tag=self.tag)
        ml_batch = safe_execute(self.processor.fetch_batch, kw)
        batch = {
            INPUT_KEY: ml_batch.input,
            LABEL_KEY: ml_batch.labels,
        }
        if ml_batch.others is not None:
            for k, v in ml_batch.others.items():
                batch[k] = v
        for k, v in self.others.items():
            batch[k] = v[item]
        batch = self.processor.postprocess_item(batch)
        return batch


class MLLoader(ArrayLoader):
    pass


@dataclass
@DataConfig.register("ml")
class MLDataConfig(DataConfig):
    batch_size: int = 128
    valid_batch_size: int = 256


@IData.register("ml")
class MLData(IData["MLData"], metaclass=ABCMeta):
    processor: MLDataProcessor
    train_dataset: MLDataset
    valid_dataset: Optional[MLDataset]

    # inheritance

    @property
    def config_base(self) -> Type[MLDataConfig]:
        return MLDataConfig

    @property
    def processor_base(self) -> Type[MLDataProcessor]:
        return MLDataProcessor

    def get_loaders(self) -> Tuple[MLLoader, Optional[MLLoader]]:
        if not self.processor.is_ready:
            raise ValueError(
                "`processor` should be ready before calling `initialize`, "
                "did you forget to call the `prepare` method first?"
            )
        if self.bundle is None:
            raise ValueError(
                "`bundle` property is not initialized, "
                "did you forget to call the `fit` method first?"
            )
        self.train_dataset = MLDataset(
            self.bundle.x_train,
            self.bundle.y_train,
            MLDatasetTag.TRAIN,
            self.processor,
            **(self.bundle.train_others or {}),
        )
        if self.bundle.x_valid is None and self.bundle.y_valid is None:
            self.valid_dataset = None
        else:
            self.valid_dataset = MLDataset(
                self.bundle.x_valid,
                self.bundle.y_valid,
                MLDatasetTag.VALID,
                self.processor,
                **(self.bundle.valid_others or {}),
            )
        train_loader = MLLoader(
            self.train_dataset,
            shuffle=self.config.shuffle_train,
            batch_size=self.config.batch_size,
            sample_weights=self.train_weights,
        )
        if self.valid_dataset is None:
            valid_loader = None
        else:
            # when `for_inference` is True, `valid_data` will always be `None`
            # so we don't need to condition `name` field here
            valid_loader = MLLoader(
                self.valid_dataset,
                shuffle=self.config.shuffle_valid,
                batch_size=self.config.valid_batch_size or self.config.batch_size,
                sample_weights=self.valid_weights,
            )
        return train_loader, valid_loader

    # api

    @property
    def num_features(self) -> int:
        if not self.processor.is_ready:
            raise ValueError("`processor` is not ready yet")
        return self.processor.get_block(GatherBlock).num_features

    @property
    def num_labels(self) -> int:
        if not self.processor.is_ready:
            raise ValueError("`processor` is not ready yet")
        return self.processor.get_block(GatherBlock).num_labels

    def build_loader(
        self,
        x: data_type,
        y: Optional[data_type] = None,
        *,
        shuffle: bool = False,
        batch_size: Optional[int] = None,
        sample_weights: sample_weights_type = None,
        **others: np.ndarray,
    ) -> MLLoader:
        self = self.copy()
        bundle = self.transform(x, y)
        x, y = bundle.x_train, bundle.y_train
        dataset = MLDataset(x, y, MLDatasetTag.TRAIN, self.processor, **others)
        loader = MLLoader(
            dataset,
            shuffle=shuffle,
            batch_size=batch_size
            or self.config.valid_batch_size
            or self.config.batch_size,
            sample_weights=sample_weights,
        )
        return loader


__all__ = [
    "MLBundledProcessorConfig",
    "MLAdvancedProcessorConfig",
    "MLDataProcessor",
    "MLDataset",
    "MLLoader",
    "MLDataConfig",
    "MLData",
]



import torch

import torch.nn as nn

from ..ops import ScaleLayer
from ..ops import SimpleInputFusion


class IHModelWithBackbone(nn.Module):
    def __init__(
        self,
        model,
        backbone,
        downsize_backbone_input=False,
        mask_fusion="sum",
        backbone_conv1_channels=64,
    ):
        """
        Creates image harmonization model supported by the features extracted from the pre-trained backbone.

        Parameters
        ----------
        model : nn.Module
            Image harmonization model takes image and mask as an input and handles features from the backbone network.
        backbone : nn.Module
            Backbone model accepts RGB image and returns a list of features.
        downsize_backbone_input : bool
            If the input image should be half-sized for the backbone.
        mask_fusion : str
            How to fuse the binary mask with the backbone input:
            'sum': apply convolution to the mask and sum it with the output of the first convolution in the backbone
            'rgb': concatenate the mask to the input image and translate it back to 3 channels with convolution
            otherwise: do not fuse mask with the backbone input
        backbone_conv1_channels : int
            If mask_fusion is 'sum', define the number of channels for the convolution applied to the mask.
        """
        super(IHModelWithBackbone, self).__init__()
        self.downsize_backbone_input = downsize_backbone_input
        self.mask_fusion = mask_fusion

        self.backbone = backbone
        self.model = model

        if mask_fusion == "rgb":
            self.fusion = SimpleInputFusion()
        elif mask_fusion == "sum":
            self.mask_conv = nn.Sequential(
                nn.Conv2d(
                    1,
                    backbone_conv1_channels,
                    kernel_size=3,
                    stride=2,
                    padding=1,
                    bias=True,
                ),
                ScaleLayer(init_value=0.1, lr_mult=1),
            )

    def forward(self, image, mask):
        """
        Forward the backbone model and then the base model, supported by the backbone feature maps.
        Return model predictions.

        Parameters
        ----------
        image : torch.Tensor
            Input RGB image.
        mask : torch.Tensor
            Binary mask of the foreground region.

        Returns
        -------
        torch.Tensor
            Harmonized RGB image.
        """
        backbone_image = image
        backbone_mask = torch.cat((mask, 1.0 - mask), dim=1)
        if self.downsize_backbone_input:
            backbone_image = nn.functional.interpolate(
                backbone_image, scale_factor=0.5, mode="bilinear", align_corners=True
            )
            backbone_mask = nn.functional.interpolate(
                backbone_mask,
                backbone_image.size()[2:],
                mode="bilinear",
                align_corners=True,
            )
        backbone_image = (
            self.fusion(backbone_image, backbone_mask[:, :1])
            if self.mask_fusion == "rgb"
            else backbone_image
        )
        backbone_mask_features = (
            self.mask_conv(backbone_mask[:, :1]) if self.mask_fusion == "sum" else None
        )
        backbone_features = self.backbone(
            backbone_image, backbone_mask, backbone_mask_features
        )

        output = self.model(image, mask, backbone_features)
        return output



import torch

import torch.nn as nn


class SimpleInputFusion(nn.Module):
    def __init__(self, add_ch=1, rgb_ch=3, ch=8, norm_layer=nn.BatchNorm2d):
        super(SimpleInputFusion, self).__init__()

        self.fusion_conv = nn.Sequential(
            nn.Conv2d(in_channels=add_ch + rgb_ch, out_channels=ch, kernel_size=1),
            nn.LeakyReLU(negative_slope=0.2),
            norm_layer(ch),
            nn.Conv2d(in_channels=ch, out_channels=rgb_ch, kernel_size=1),
        )

    def forward(self, image, additional_input):
        return self.fusion_conv(torch.cat((image, additional_input), dim=1))


class ChannelAttention(nn.Module):
    def __init__(self, in_channels):
        super(ChannelAttention, self).__init__()
        self.global_pools = nn.ModuleList(
            [
                nn.AdaptiveAvgPool2d(1),
                nn.AdaptiveMaxPool2d(1),
            ]
        )
        intermediate_channels_count = max(in_channels // 16, 8)
        self.attention_transform = nn.Sequential(
            nn.Linear(
                len(self.global_pools) * in_channels, intermediate_channels_count
            ),
            nn.ReLU(),
            nn.Linear(intermediate_channels_count, in_channels),
            nn.Sigmoid(),
        )

    def forward(self, x):
        pooled_x = []
        for global_pool in self.global_pools:
            pooled_x.append(global_pool(x))
        pooled_x = torch.cat(pooled_x, dim=1).flatten(start_dim=1)
        channel_attention_weights = self.attention_transform(pooled_x)[..., None, None]
        return channel_attention_weights * x


class MaskedChannelAttention(nn.Module):
    def __init__(self, in_channels, *args, **kwargs):
        super(MaskedChannelAttention, self).__init__()
        self.global_max_pool = MaskedGlobalMaxPool2d()
        self.global_avg_pool = FastGlobalAvgPool2d()

        intermediate_channels_count = max(in_channels // 16, 8)
        self.attention_transform = nn.Sequential(
            nn.Linear(3 * in_channels, intermediate_channels_count),
            nn.ReLU(inplace=True),
            nn.Linear(intermediate_channels_count, in_channels),
            nn.Sigmoid(),
        )

    def forward(self, x, mask):
        if mask.shape[2:] != x.shape[:2]:
            mask = nn.functional.interpolate(
                mask, size=x.size()[-2:], mode="bilinear", align_corners=True
            )
        pooled_x = torch.cat(
            [self.global_max_pool(x, mask), self.global_avg_pool(x)], dim=1
        )
        channel_attention_weights = self.attention_transform(pooled_x)[..., None, None]

        return channel_attention_weights * x


class MaskedGlobalMaxPool2d(nn.Module):
    def __init__(self):
        super().__init__()
        self.global_max_pool = FastGlobalMaxPool2d()

    def forward(self, x, mask):
        return torch.cat(
            (self.global_max_pool(x * mask), self.global_max_pool(x * (1.0 - mask))),
            dim=1,
        )


class FastGlobalAvgPool2d(nn.Module):
    def __init__(self):
        super(FastGlobalAvgPool2d, self).__init__()

    def forward(self, x):
        in_size = x.size()
        return x.view((in_size[0], in_size[1], -1)).mean(dim=2)


class FastGlobalMaxPool2d(nn.Module):
    def __init__(self):
        super(FastGlobalMaxPool2d, self).__init__()

    def forward(self, x):
        in_size = x.size()
        return x.view((in_size[0], in_size[1], -1)).max(dim=2)[0]


class ScaleLayer(nn.Module):
    def __init__(self, init_value=1.0, lr_mult=1):
        super().__init__()
        self.lr_mult = lr_mult
        self.scale = nn.Parameter(
            torch.full((1,), init_value / lr_mult, dtype=torch.float32)
        )

    def forward(self, x):
        scale = torch.abs(self.scale * self.lr_mult)
        return x * scale


class FeaturesConnector(nn.Module):
    def __init__(self, mode, in_channels, feature_channels, out_channels):
        super(FeaturesConnector, self).__init__()
        self.mode = mode if feature_channels else ""

        if self.mode == "catc":
            self.reduce_conv = nn.Conv2d(
                in_channels + feature_channels, out_channels, kernel_size=1
            )
        elif self.mode == "sum":
            self.reduce_conv = nn.Conv2d(feature_channels, out_channels, kernel_size=1)

        self.output_channels = (
            out_channels if self.mode != "cat" else in_channels + feature_channels
        )

    def forward(self, x, features):
        if self.mode == "cat":
            return torch.cat((x, features), 1)
        if self.mode == "catc":
            return self.reduce_conv(torch.cat((x, features), 1))
        if self.mode == "sum":
            return self.reduce_conv(features) + x
        return x

    def extra_repr(self):
        return self.mode



import torch.nn as nn

from torch import Tensor
from typing import Any
from torch.nn import Module


class Residual(Module):
    def __init__(self, module: Module):
        super().__init__()
        self.module = module

    def forward(self, net: Tensor, **kwargs: Any) -> Tensor:
        return net + self.module(net, **kwargs)


def zero_module(module: Module) -> Module:
    for p in module.parameters():
        p.detach().zero_()
    return module


def avg_pool_nd(n: int, *args: Any, **kwargs: Any) -> Module:
    if n == 1:
        return nn.AvgPool1d(*args, **kwargs)
    elif n == 2:
        return nn.AvgPool2d(*args, **kwargs)
    elif n == 3:
        return nn.AvgPool3d(*args, **kwargs)
    raise ValueError(f"unsupported dimensions: {n}")


__all__ = [
    "Residual",
    "avg_pool_nd",
    "zero_module",
]



# Dependency Hierarchy
## utils, common, customs, hooks, norms
### hijacks (hooks, customs)
### activations (common)
#### mappings (norms, customs, activations)
#### convs.basic (norms, hijacks, activations)
#### convs.residual (convs.basic, utils)
##### attentions (convs, common, customs, activations)
###### high_level (all the above)
####### mixed_stacks (all the above)
######## implementations (all the above)

from .utils import *
from .common import *
from .customs import *
from .hooks import *
from .norms import *
from .hijacks import *
from .activations import *
from .mappings import *
from .convs import *
from .attentions import *
from .high_level import *
from .mixed_stacks import *
from .implementations import *



from ..model.base import ISEUNetV1
from ..model.base import DeepImageHarmonization
from ..model.base import SSAMImageHarmonization


BMCONFIGS = {
    "dih256": {"model": DeepImageHarmonization, "params": {"depth": 7}},
    "improved_dih256": {
        "model": DeepImageHarmonization,
        "params": {"depth": 7, "batchnorm_from": 2, "image_fusion": True},
    },
    "improved_sedih256": {
        "model": DeepImageHarmonization,
        "params": {
            "depth": 7,
            "batchnorm_from": 2,
            "image_fusion": True,
            "attend_from": 5,
        },
    },
    "ssam256": {
        "model": SSAMImageHarmonization,
        "params": {"depth": 4, "batchnorm_from": 2, "attend_from": 2},
    },
    "improved_ssam256": {
        "model": SSAMImageHarmonization,
        "params": {
            "depth": 4,
            "ch": 32,
            "image_fusion": True,
            "attention_mid_k": 0.5,
            "batchnorm_from": 2,
            "attend_from": 2,
        },
    },
    "iseunetv1_256": {
        "model": ISEUNetV1,
        "params": {"depth": 4, "batchnorm_from": 2, "attend_from": 1, "ch": 64},
    },
    "dih512": {"model": DeepImageHarmonization, "params": {"depth": 8}},
    "improved_dih512": {
        "model": DeepImageHarmonization,
        "params": {"depth": 8, "batchnorm_from": 2, "image_fusion": True},
    },
    "improved_ssam512": {
        "model": SSAMImageHarmonization,
        "params": {
            "depth": 6,
            "ch": 32,
            "image_fusion": True,
            "attention_mid_k": 0.5,
            "batchnorm_from": 2,
            "attend_from": 3,
        },
    },
    "improved_sedih512": {
        "model": DeepImageHarmonization,
        "params": {
            "depth": 8,
            "batchnorm_from": 2,
            "image_fusion": True,
            "attend_from": 6,
        },
    },
}



import numpy as np
import math
import torch

from . import util
from .model import bodypose_model

try:
    import cv2
except:
    cv2 = None
try:
    from scipy.ndimage.filters import gaussian_filter
except:
    gaussian_filter = None
try:
    import matplotlib.pyplot as plt
except:
    plt = None


class Body:
    def __init__(self, model_path, device="cpu"):
        if cv2 is None:
            raise ValueError("`cv2` is needed for `Body`")
        if gaussian_filter is None:
            raise ValueError("`scipy` is needed for `Body`")
        if plt is None:
            raise ValueError("`matplotlib` is needed for `Body`")
        self.model = bodypose_model()
        model_dict = util.transfer(self.model, torch.load(model_path))
        self.model.load_state_dict(model_dict)
        self.model.eval()
        self.model.to(device)

    def __call__(self, oriImg, to):
        # scale_search = [0.5, 1.0, 1.5, 2.0]
        scale_search = [0.5]
        boxsize = 368
        stride = 8
        padValue = 128
        thre1 = 0.1
        thre2 = 0.05
        multiplier = [x * boxsize / oriImg.shape[0] for x in scale_search]
        heatmap_avg = np.zeros((oriImg.shape[0], oriImg.shape[1], 19))
        paf_avg = np.zeros((oriImg.shape[0], oriImg.shape[1], 38))

        for m in range(len(multiplier)):
            scale = multiplier[m]
            imageToTest = cv2.resize(
                oriImg, (0, 0), fx=scale, fy=scale, interpolation=cv2.INTER_CUBIC
            )
            imageToTest_padded, pad = util.padRightDownCorner(
                imageToTest, stride, padValue
            )
            im = (
                np.transpose(
                    np.float32(imageToTest_padded[:, :, :, np.newaxis]), (3, 2, 0, 1)
                )
                / 256
                - 0.5
            )
            im = np.ascontiguousarray(im)

            data = to(torch.from_numpy(im))
            # data = data.permute([2, 0, 1]).unsqueeze(0).float()
            with torch.no_grad():
                Mconv7_stage6_L1, Mconv7_stage6_L2 = self.model(data)
            Mconv7_stage6_L1 = Mconv7_stage6_L1.cpu().float().numpy()
            Mconv7_stage6_L2 = Mconv7_stage6_L2.cpu().float().numpy()

            # extract outputs, resize, and remove padding
            # heatmap = np.transpose(np.squeeze(net.blobs[output_blobs.keys()[1]].data), (1, 2, 0))  # output 1 is heatmaps
            heatmap = np.transpose(
                np.squeeze(Mconv7_stage6_L2), (1, 2, 0)
            )  # output 1 is heatmaps
            heatmap = cv2.resize(
                heatmap, (0, 0), fx=stride, fy=stride, interpolation=cv2.INTER_CUBIC
            )
            heatmap = heatmap[
                : imageToTest_padded.shape[0] - pad[2],
                : imageToTest_padded.shape[1] - pad[3],
                :,
            ]
            heatmap = cv2.resize(
                heatmap,
                (oriImg.shape[1], oriImg.shape[0]),
                interpolation=cv2.INTER_CUBIC,
            )

            # paf = np.transpose(np.squeeze(net.blobs[output_blobs.keys()[0]].data), (1, 2, 0))  # output 0 is PAFs
            paf = np.transpose(
                np.squeeze(Mconv7_stage6_L1), (1, 2, 0)
            )  # output 0 is PAFs
            paf = cv2.resize(
                paf, (0, 0), fx=stride, fy=stride, interpolation=cv2.INTER_CUBIC
            )
            paf = paf[
                : imageToTest_padded.shape[0] - pad[2],
                : imageToTest_padded.shape[1] - pad[3],
                :,
            ]
            paf = cv2.resize(
                paf, (oriImg.shape[1], oriImg.shape[0]), interpolation=cv2.INTER_CUBIC
            )

            heatmap_avg += heatmap_avg + heatmap / len(multiplier)
            paf_avg += +paf / len(multiplier)

        all_peaks = []
        peak_counter = 0

        for part in range(18):
            map_ori = heatmap_avg[:, :, part]
            one_heatmap = gaussian_filter(map_ori, sigma=3)

            map_left = np.zeros(one_heatmap.shape)
            map_left[1:, :] = one_heatmap[:-1, :]
            map_right = np.zeros(one_heatmap.shape)
            map_right[:-1, :] = one_heatmap[1:, :]
            map_up = np.zeros(one_heatmap.shape)
            map_up[:, 1:] = one_heatmap[:, :-1]
            map_down = np.zeros(one_heatmap.shape)
            map_down[:, :-1] = one_heatmap[:, 1:]

            peaks_binary = np.logical_and.reduce(
                (
                    one_heatmap >= map_left,
                    one_heatmap >= map_right,
                    one_heatmap >= map_up,
                    one_heatmap >= map_down,
                    one_heatmap > thre1,
                )
            )
            peaks = list(
                zip(np.nonzero(peaks_binary)[1], np.nonzero(peaks_binary)[0])
            )  # note reverse
            peaks_with_score = [x + (map_ori[x[1], x[0]],) for x in peaks]
            peak_id = range(peak_counter, peak_counter + len(peaks))
            peaks_with_score_and_id = [
                peaks_with_score[i] + (peak_id[i],) for i in range(len(peak_id))
            ]

            all_peaks.append(peaks_with_score_and_id)
            peak_counter += len(peaks)

        # find connection in the specified sequence, center 29 is in the position 15
        limbSeq = [
            [2, 3],
            [2, 6],
            [3, 4],
            [4, 5],
            [6, 7],
            [7, 8],
            [2, 9],
            [9, 10],
            [10, 11],
            [2, 12],
            [12, 13],
            [13, 14],
            [2, 1],
            [1, 15],
            [15, 17],
            [1, 16],
            [16, 18],
            [3, 17],
            [6, 18],
        ]
        # the middle joints heatmap correpondence
        mapIdx = [
            [31, 32],
            [39, 40],
            [33, 34],
            [35, 36],
            [41, 42],
            [43, 44],
            [19, 20],
            [21, 22],
            [23, 24],
            [25, 26],
            [27, 28],
            [29, 30],
            [47, 48],
            [49, 50],
            [53, 54],
            [51, 52],
            [55, 56],
            [37, 38],
            [45, 46],
        ]

        connection_all = []
        special_k = []
        mid_num = 10

        for k in range(len(mapIdx)):
            score_mid = paf_avg[:, :, [x - 19 for x in mapIdx[k]]]
            candA = all_peaks[limbSeq[k][0] - 1]
            candB = all_peaks[limbSeq[k][1] - 1]
            nA = len(candA)
            nB = len(candB)
            indexA, indexB = limbSeq[k]
            if nA != 0 and nB != 0:
                connection_candidate = []
                for i in range(nA):
                    for j in range(nB):
                        vec = np.subtract(candB[j][:2], candA[i][:2])
                        norm = math.sqrt(vec[0] * vec[0] + vec[1] * vec[1])
                        norm = max(0.001, norm)
                        vec = np.divide(vec, norm)

                        startend = list(
                            zip(
                                np.linspace(candA[i][0], candB[j][0], num=mid_num),
                                np.linspace(candA[i][1], candB[j][1], num=mid_num),
                            )
                        )

                        vec_x = np.array(
                            [
                                score_mid[
                                    int(round(startend[I][1])),
                                    int(round(startend[I][0])),
                                    0,
                                ]
                                for I in range(len(startend))
                            ]
                        )
                        vec_y = np.array(
                            [
                                score_mid[
                                    int(round(startend[I][1])),
                                    int(round(startend[I][0])),
                                    1,
                                ]
                                for I in range(len(startend))
                            ]
                        )

                        score_midpts = np.multiply(vec_x, vec[0]) + np.multiply(
                            vec_y, vec[1]
                        )
                        score_with_dist_prior = sum(score_midpts) / len(
                            score_midpts
                        ) + min(0.5 * oriImg.shape[0] / norm - 1, 0)
                        criterion1 = len(
                            np.nonzero(score_midpts > thre2)[0]
                        ) > 0.8 * len(score_midpts)
                        criterion2 = score_with_dist_prior > 0
                        if criterion1 and criterion2:
                            connection_candidate.append(
                                [
                                    i,
                                    j,
                                    score_with_dist_prior,
                                    score_with_dist_prior + candA[i][2] + candB[j][2],
                                ]
                            )

                connection_candidate = sorted(
                    connection_candidate, key=lambda x: x[2], reverse=True
                )
                connection = np.zeros((0, 5))
                for c in range(len(connection_candidate)):
                    i, j, s = connection_candidate[c][0:3]
                    if i not in connection[:, 3] and j not in connection[:, 4]:
                        connection = np.vstack(
                            [connection, [candA[i][3], candB[j][3], s, i, j]]
                        )
                        if len(connection) >= min(nA, nB):
                            break

                connection_all.append(connection)
            else:
                special_k.append(k)
                connection_all.append([])

        # last number in each row is the total parts number of that person
        # the second last number in each row is the score of the overall configuration
        subset = -1 * np.ones((0, 20))
        candidate = np.array([item for sublist in all_peaks for item in sublist])

        for k in range(len(mapIdx)):
            if k not in special_k:
                partAs = connection_all[k][:, 0]
                partBs = connection_all[k][:, 1]
                indexA, indexB = np.array(limbSeq[k]) - 1

                for i in range(len(connection_all[k])):  # = 1:size(temp,1)
                    found = 0
                    subset_idx = [-1, -1]
                    for j in range(len(subset)):  # 1:size(subset,1):
                        if (
                            subset[j][indexA] == partAs[i]
                            or subset[j][indexB] == partBs[i]
                        ):
                            subset_idx[found] = j
                            found += 1

                    if found == 1:
                        j = subset_idx[0]
                        if subset[j][indexB] != partBs[i]:
                            subset[j][indexB] = partBs[i]
                            subset[j][-1] += 1
                            subset[j][-2] += (
                                candidate[partBs[i].astype(int), 2]
                                + connection_all[k][i][2]
                            )
                    elif found == 2:  # if found 2 and disjoint, merge them
                        j1, j2 = subset_idx
                        membership = (
                            (subset[j1] >= 0).astype(int)
                            + (subset[j2] >= 0).astype(int)
                        )[:-2]
                        if len(np.nonzero(membership == 2)[0]) == 0:  # merge
                            subset[j1][:-2] += subset[j2][:-2] + 1
                            subset[j1][-2:] += subset[j2][-2:]
                            subset[j1][-2] += connection_all[k][i][2]
                            subset = np.delete(subset, j2, 0)
                        else:  # as like found == 1
                            subset[j1][indexB] = partBs[i]
                            subset[j1][-1] += 1
                            subset[j1][-2] += (
                                candidate[partBs[i].astype(int), 2]
                                + connection_all[k][i][2]
                            )

                    # if find no partA in the subset, create a new subset
                    elif not found and k < 17:
                        row = -1 * np.ones(20)
                        row[indexA] = partAs[i]
                        row[indexB] = partBs[i]
                        row[-1] = 2
                        row[-2] = (
                            sum(candidate[connection_all[k][i, :2].astype(int), 2])
                            + connection_all[k][i][2]
                        )
                        subset = np.vstack([subset, row])
        # delete some rows of subset which has few parts occur
        deleteIdx = []
        for i in range(len(subset)):
            if subset[i][-1] < 4 or subset[i][-2] / subset[i][-1] < 0.4:
                deleteIdx.append(i)
        subset = np.delete(subset, deleteIdx, axis=0)

        # subset: n*20 array, 0-17 is the index in candidate, 18 is the total score, 19 is the total parts
        # candidate: x, y, score, id
        return candidate, subset


if __name__ == "__main__":
    body_estimation = Body("../model/body_pose_model.pth")

    test_image = "../images/ski.jpg"
    oriImg = cv2.imread(test_image)  # B,G,R order
    candidate, subset = body_estimation(oriImg)
    canvas = util.draw_bodypose(oriImg, candidate, subset)
    plt.imshow(canvas[:, :, [2, 1, 0]])
    plt.show()



import math
import torch

import numpy as np
import torch.nn as nn
import torch.nn.functional as F

from torch import Tensor
from typing import Any
from typing import Dict
from typing import List
from typing import Tuple
from typing import Union
from typing import Callable
from typing import Optional
from functools import partial
from torch.nn import Module
from cftool.misc import print_warning

from .hooks import IBasicHook


class Linear(Module):
    def __init__(
        self,
        in_dim: int,
        out_dim: int,
        *,
        bias: bool = True,
        pruner_config: Optional[Dict[str, Any]] = None,
        init_method: Optional[str] = None,
        rank: Optional[int] = None,
        rank_ratio: Optional[float] = None,
        hook: Optional[IBasicHook] = None,
    ):
        super().__init__()
        original_rank = min(in_dim, out_dim)
        if rank is None:
            if rank_ratio is not None:
                rank = round(original_rank * rank_ratio)
        if rank is not None and rank >= original_rank:
            print_warning(
                f"specified rank ({rank}) should be smaller than "
                f"the original rank ({original_rank})"
            )
        if rank is None:
            self.w1 = self.w2 = self.b = None
            self.linear = nn.Linear(in_dim, out_dim, bias)
        else:
            self.w1 = nn.Parameter(torch.zeros(rank, in_dim))
            self.w2 = nn.Parameter(torch.zeros(out_dim, rank))
            self.b = nn.Parameter(torch.zeros(1, out_dim)) if bias else None
            self.linear = None
        if pruner_config is None:
            self.pruner = self.pruner1 = self.pruner2 = None
        else:
            if rank is None:
                self.pruner1 = self.pruner2 = None
                self.pruner = Pruner(pruner_config, [out_dim, in_dim])
            else:
                self.pruner1 = Pruner(pruner_config, [rank, in_dim])
                self.pruner2 = Pruner(pruner_config, [out_dim, rank])
                self.pruner = None
        # initialize
        init_method = init_method or "xavier_normal"
        init_fn = getattr(nn.init, f"{init_method}_", nn.init.xavier_normal_)
        self.init_weights_with(lambda t: init_fn(t, 1.0 / math.sqrt(2.0)))
        # hook
        self.hook = hook

    @property
    def weight(self) -> Tensor:
        if self.linear is not None:
            return self.linear.weight
        return torch.matmul(self.w2, self.w1)

    @property
    def bias(self) -> Optional[Tensor]:
        return self.b if self.linear is None else self.linear.bias

    def forward(self, net: Tensor) -> Tensor:
        inp = net
        if self.hook is not None:
            inp = self.hook.before_forward(inp)
        if self.linear is not None:
            weight = self.linear.weight
            if self.pruner is not None:
                weight = self.pruner(weight)
            net = F.linear(net, weight, self.linear.bias)
        else:
            w1, w2 = self.w1, self.w2
            if self.pruner1 is not None:
                w1 = self.pruner1(w1)
            if self.pruner2 is not None:
                w2 = self.pruner2(w2)
            net = F.linear(net, w1)
            net = F.linear(net, w2, self.b)
        if self.hook is not None:
            net = self.hook.after_forward(inp, net)
        return net

    def init_weights_with(self, w_init_fn: Callable[[Tensor], None]) -> None:
        with torch.no_grad():
            if self.linear is not None:
                w_init_fn(self.linear.weight.data)
            elif self.w1 is not None and self.w2 is not None:
                w_init_fn(self.w1.data)
                w_init_fn(self.w2.data)
            if self.linear is None:
                if self.b is not None:
                    self.b.data.zero_()
            else:
                if self.linear.bias is not None:
                    self.linear.bias.data.zero_()


class LeafAggregation(torch.autograd.Function):
    @staticmethod
    def forward(ctx: Any, *args: Any, **kwargs: Any) -> Tensor:
        net, leaves = args
        softmax_leaves = F.softmax(leaves, dim=1)
        ctx.save_for_backward(net, softmax_leaves.t())
        return net.mm(softmax_leaves)

    @staticmethod
    def backward(ctx: Any, *grad_outputs: Any) -> Tuple[Optional[Tensor], ...]:
        grad_output = grad_outputs[0]
        if grad_output is None:
            return None, None
        net, softmax_leaves = ctx.saved_tensors
        net_grad = grad_output.mm(softmax_leaves)
        sub_grad = grad_output.t().mm(net)
        sub_grad2 = (softmax_leaves * sub_grad).sum(0, keepdim=True)
        leaves_grad = softmax_leaves * (sub_grad - sub_grad2)
        return net_grad, leaves_grad.t()


class Route(torch.autograd.Function):
    @staticmethod
    def forward(ctx: Any, *args: Any, **kwargs: Any) -> Tensor:
        (
            net,
            tree_arange,
            batch_indices,
            ones,
            increment_masks,
            num_tree,
            num_batch,
            tree_depth,
            num_internals,
        ) = args
        shape = num_batch, -1, num_internals
        sigmoid_net = torch.sigmoid(net)
        p_left = sigmoid_net.view(*shape).transpose(0, 1)
        p_right = 1.0 - p_left
        flat_probabilities = torch.cat([p_left, p_right], dim=-1)
        flat_probabilities = flat_probabilities.contiguous().view(num_tree, -1)
        current_indices = batch_indices + increment_masks[0]
        flat_dim = flat_probabilities.shape[-1]
        tree_arange = tree_arange * flat_dim
        routes = flat_probabilities.take(tree_arange + current_indices[None, ...])
        all_routes = [routes.clone()]
        for i in range(1, tree_depth + 1):
            current_indices = batch_indices + increment_masks[i]
            current_indices = tree_arange + current_indices[None, ...]
            current_routes = flat_probabilities.take(current_indices)
            all_routes.append(current_routes)
            routes *= current_routes
        ctx.save_for_backward(ones, sigmoid_net, *all_routes)
        ctx.tree_depth = tree_depth
        ctx.num_tree = num_tree
        return routes

    @staticmethod
    def backward(ctx: Any, *grad_outputs: Any) -> Tuple[Optional[Tensor], ...]:
        grad_output = grad_outputs[0]
        dummy_grads = tuple(None for _ in range(8))
        if grad_output is None:
            return (None,) + dummy_grads
        num_tree = ctx.num_tree
        tree_depth = ctx.tree_depth
        ones_list, sigmoid_net, *all_routes = ctx.saved_tensors
        cursor = 0
        divide = 1
        num_leaves = 2 ** (tree_depth + 1)
        sub_grads_shape = num_tree, sigmoid_net.shape[0], num_leaves - 1
        sub_grads = torch.zeros(*sub_grads_shape, device=grad_output.device)
        for i in range(tree_depth + 1):
            ones = ones_list[i]
            nodes = ones[None, None, ...]
            for j in range(tree_depth + 1):
                if j == i:
                    continue
                nodes = nodes * all_routes[j]
            sub_grad = grad_output * nodes
            section = int(round(num_leaves / divide))
            sub_grad = sub_grad.view(num_tree, -1, divide, section)
            sub_grad = sub_grad.sum(-1)
            sub_grads[..., cursor : cursor + divide] = sub_grad
            cursor += divide
            divide *= 2

        sub_grads = sub_grads.transpose(0, 1).contiguous()
        sub_grads = sub_grads.view(-1, num_tree * (num_leaves - 1))
        return (sigmoid_net * (1.0 - sigmoid_net) * sub_grads,) + dummy_grads


class DNDF(Module):
    def __init__(
        self,
        in_dim: int,
        out_dim: Optional[int],
        *,
        num_tree: int = 10,
        tree_depth: int = 4,
        is_regression: Optional[bool] = None,
        tree_proj_config: Optional[Dict[str, Any]] = None,
        use_fast_dndf: bool = True,
    ):
        super().__init__()
        self._num_tree = num_tree
        self._tree_depth = tree_depth
        self._is_regression = is_regression
        if out_dim is not None and is_regression is None:
            self._is_regression = out_dim == 1
        self._num_leaf = 2 ** (self._tree_depth + 1)
        self._num_internals = self._num_leaf - 1
        self._output_dim = out_dim
        self._fast = use_fast_dndf
        if tree_proj_config is None:
            tree_proj_config = {}
        tree_proj_config.setdefault("pruner_config", {})
        self.tree_proj = Linear(
            in_dim,
            self._num_internals * self._num_tree,
            **tree_proj_config,
        )
        if out_dim is None:
            self.leaves = None
        else:
            leaves_shape = self._num_tree * self._num_leaf, out_dim
            self.leaves = nn.Parameter(torch.empty(*leaves_shape))
            with torch.no_grad():
                torch.nn.init.xavier_uniform_(self.leaves.data)
        # buffers
        num_repeat, num_local_internals = self._num_leaf // 2, 1
        ones_np = np.repeat([1, -1], num_repeat)
        ones_list = [torch.from_numpy(ones_np.astype(np.float32))]
        increment_indices_np = np.repeat([0, self._num_internals], num_repeat)
        increment_indices = [torch.from_numpy(increment_indices_np.astype(np.int64))]
        for i in range(1, self._tree_depth + 1):
            num_repeat //= 2
            num_local_internals *= 2
            arange = np.arange(num_local_internals - 1, 2 * num_local_internals - 1)
            ones_np = np.repeat([1, -1], num_repeat)
            ones_np = np.tile(ones_np, 2**i)
            ones_list.append(torch.from_numpy(ones_np.astype(np.float32)))
            increment_mask = np.repeat(arange, 2)
            increment_mask += np.tile([0, self._num_internals], num_local_internals)
            increment_mask = np.repeat(increment_mask, num_repeat)
            increment_mask_ = torch.from_numpy(increment_mask.astype(np.int64))
            increment_indices.append(increment_mask_)
        self.increment_masks: Tensor
        self.register_buffer("tree_arange", torch.arange(num_tree)[..., None, None])
        self.register_buffer("ones", torch.stack(ones_list))
        self.register_buffer("increment_indices", torch.stack(increment_indices))

    def forward(self, net: Tensor) -> Tensor:
        num_batch = net.shape[0]
        tree_net = self.tree_proj(net)

        num_flat_prob = 2 * self._num_internals
        arange_args = 0, num_flat_prob * num_batch, num_flat_prob
        batch_indices = torch.arange(*arange_args, device=tree_net.device).view(-1, 1)

        if self._fast:
            routes = Route.apply(
                tree_net,
                self.tree_arange,
                batch_indices,
                self.ones,
                self.increment_indices,
                self._num_tree,
                num_batch,
                self._tree_depth,
                self._num_internals,
            )
        else:
            shape = num_batch, -1, self._num_internals
            p_left = torch.sigmoid(tree_net).view(*shape).transpose(0, 1)
            p_right = 1.0 - p_left
            flat_probabilities = torch.cat([p_left, p_right], dim=-1).contiguous()
            flat_probabilities = flat_probabilities.view(self._num_tree, -1)
            current_indices = batch_indices + self.increment_indices[0]  # type: ignore
            flat_dim = flat_probabilities.shape[-1]
            tree_arange = self.tree_arange * flat_dim  # type: ignore
            routes = flat_probabilities.take(tree_arange + current_indices[None, ...])
            for i in range(1, self._tree_depth + 1):
                current_indices = batch_indices + self.increment_indices[i]  # type: ignore
                current_indices = tree_arange + current_indices[None, ...]
                routes *= flat_probabilities.take(current_indices)

        features = routes.transpose(0, 1).contiguous().view(num_batch, -1)
        if self.leaves is None or self._output_dim is None:
            return features.view(num_batch, self._num_tree, -1)
        if self._is_regression or self._output_dim <= 1:
            outputs = features.mm(self.leaves)
        else:
            if self._fast:
                outputs = LeafAggregation.apply(features, self.leaves)
            else:
                leaves = F.softmax(self.leaves, dim=1)
                outputs = features.mm(leaves)
        return outputs / self._num_tree


class Pruner(Module):
    def __init__(self, config: Dict[str, Any], w_shape: Optional[List[int]] = None):
        super().__init__()
        self.eps: Tensor
        self.exp: Tensor
        self.alpha: Union[Tensor, nn.Parameter]
        self.beta: Union[Tensor, nn.Parameter]
        self.gamma: Union[Tensor, nn.Parameter]
        self.max_ratio: Union[Tensor, nn.Parameter]
        tensor = partial(torch.tensor, dtype=torch.float32)
        self.method = config.setdefault("method", "auto_prune")
        if self.method == "surgery":
            if w_shape is None:
                msg = "`w_shape` of `Pruner` should be provided when `surgery` is used"
                raise ValueError(msg)
            self.register_buffer("mask", torch.ones(*w_shape, dtype=torch.float32))
            self.register_buffer("alpha", tensor([config.setdefault("alpha", 1.0)]))
            self.register_buffer("beta", tensor([config.setdefault("beta", 4.0)]))
            self.register_buffer("gamma", tensor([config.setdefault("gamma", 1e-4)]))
            self.register_buffer("eps", tensor([config.setdefault("eps", 1e-12)]))
            keys = ["alpha", "beta", "gamma", "eps"]
        elif self.method == "simplified":
            self.register_buffer("alpha", tensor([config.setdefault("alpha", 0.01)]))
            self.register_buffer("beta", tensor([config.setdefault("beta", 1.0)]))
            self.register_buffer(
                "max_ratio", tensor([config.setdefault("max_ratio", 1.0)])
            )
            self.register_buffer("exp", tensor([config.setdefault("exp", 0.5)]))
            keys = ["alpha", "beta", "max_ratio", "exp"]
        else:
            self.register_buffer(
                "alpha",
                tensor(
                    [
                        config.setdefault(
                            "alpha", 1e-4 if self.method == "hard_prune" else 1e-2
                        )
                    ]
                ),
            )
            self.register_buffer("beta", tensor([config.setdefault("beta", 1.0)]))
            self.register_buffer("gamma", tensor([config.setdefault("gamma", 1.0)]))
            self.register_buffer(
                "max_ratio", tensor([config.setdefault("max_ratio", 1.0)])
            )
            if not all(
                scalar > 0
                for scalar in [self.alpha, self.beta, self.gamma, self.max_ratio]
            ):
                raise ValueError("parameters should greater than 0. in pruner")
            self.register_buffer("eps", tensor([config.setdefault("eps", 1e-12)]))
            if self.method == "auto_prune":
                for attr in ["alpha", "beta", "gamma", "max_ratio"]:
                    setattr(self, attr, torch.log(torch.exp(getattr(self, attr)) - 1))
                self.alpha, self.beta, self.gamma, self.max_ratio = map(
                    lambda param: nn.Parameter(param),
                    [self.alpha, self.beta, self.gamma, self.max_ratio],
                )
            keys = ["alpha", "beta", "gamma", "max_ratio", "eps"]
        self._repr_keys = keys

    def forward(self, w: Tensor) -> Tensor:
        w_abs = torch.abs(w)
        if self.method == "surgery":
            mu, std = torch.mean(w_abs), torch.std(w_abs)
            zeros_mask = self.mask == 0.0
            ones_mask = self.mask == 1.0
            to_zeros_mask = ones_mask & (w_abs <= 0.9 * (mu - self.beta * std))  # type: ignore
            to_ones_mask = zeros_mask & (w_abs >= 1.1 * (mu + self.beta * std))  # type: ignore
            self.mask.masked_fill(to_zeros_mask, 0.0)  # type: ignore
            self.mask.masked_fill(to_ones_mask, 1.0)  # type: ignore
            mask = self.mask
            del mu, std, ones_mask, zeros_mask, to_zeros_mask, to_ones_mask
        else:
            if self.method != "auto_prune":
                alpha, beta, ratio = self.alpha, self.beta, self.max_ratio
            else:
                alpha, beta, ratio = map(
                    F.softplus,
                    [self.alpha, self.beta, self.max_ratio],
                )
            if self.method == "simplified":
                log_w = torch.min(ratio, beta * torch.pow(w_abs, self.exp))
            else:
                w_abs_mean = torch.mean(w_abs)
                if self.method != "auto_prune":
                    gamma = self.gamma
                else:
                    gamma = F.softplus(self.gamma)
                log_w = torch.log(torch.max(self.eps, w_abs / (w_abs_mean * gamma)))
                log_w = torch.min(ratio, beta * log_w)
                del w_abs_mean
            mask = torch.max(alpha / beta * log_w, log_w)
            del log_w
        w = w * mask
        del w_abs, mask
        return w

    def extra_repr(self) -> str:
        if self.method == "auto_prune":
            return f"method='{self.method}'"
        max_str_len = max(map(len, self._repr_keys))
        return "\n".join(
            [f"(0): method={self.method}\n(1): Settings("]
            + [
                f"  {key:<{max_str_len}s} - {getattr(self, key).item()}"
                for key in self._repr_keys
            ]
            + [")"]
        )


class DropPath(Module):
    def __init__(self, dropout: float = 0.0):
        super().__init__()
        self.dropout = dropout

    def forward(self, net: Tensor) -> Tensor:
        if not 0.0 < self.dropout < 1.0 or not self.training:
            return net
        keep_prob = 1.0 - self.dropout
        shape = (net.shape[0],) + (1,) * (net.ndim - 1)
        rand = torch.rand(shape, dtype=net.dtype, device=net.device)
        random_tensor = keep_prob + rand
        random_tensor.floor_()
        net = net.div(keep_prob) * random_tensor
        return net

    def extra_repr(self) -> str:
        return str(self.dropout)


__all__ = [
    "Linear",
    "LeafAggregation",
    "Route",
    "DNDF",
    "Pruner",
    "DropPath",
]



import numpy as np

from cftool.types import np_dict_type

from ....schema import IRuntimeDataBlock
from ....constants import INPUT_KEY


@IRuntimeDataBlock.register("hwc_to_chw")
class HWCToCHWBlock(IRuntimeDataBlock):
    def postprocess_item(self, item: np_dict_type) -> np_dict_type:
        inp = item[INPUT_KEY].transpose([2, 0, 1])
        inp = np.ascontiguousarray(inp)
        item[INPUT_KEY] = inp
        return item


__all__ = ["HWCToCHWBlock"]



import torch

import torch.nn as nn

from .basic_blocks import ConvBlock
from ..ops import FeaturesConnector
from ..ops import MaskedChannelAttention


class ConvEncoder(nn.Module):
    def __init__(
        self,
        depth,
        ch,
        norm_layer,
        batchnorm_from,
        max_channels,
        backbone_from,
        backbone_channels=None,
        backbone_mode="",
    ):
        super(ConvEncoder, self).__init__()
        self.depth = depth
        self.backbone_from = backbone_from
        backbone_channels = [] if backbone_channels is None else backbone_channels[::-1]

        in_channels = 4
        out_channels = ch

        self.block0 = ConvBlock(
            in_channels,
            out_channels,
            norm_layer=norm_layer if batchnorm_from == 0 else None,
        )
        self.block1 = ConvBlock(
            out_channels,
            out_channels,
            norm_layer=norm_layer if 0 <= batchnorm_from <= 1 else None,
        )
        self.blocks_channels = [out_channels, out_channels]

        self.blocks_connected = nn.ModuleDict()
        self.connectors = nn.ModuleDict()
        for block_i in range(2, depth):
            if block_i % 2:
                in_channels = out_channels
            else:
                in_channels, out_channels = out_channels, min(
                    2 * out_channels, max_channels
                )

            if 0 <= backbone_from <= block_i and len(backbone_channels):
                stage_channels = backbone_channels.pop()
                connector = FeaturesConnector(
                    backbone_mode, in_channels, stage_channels, in_channels
                )
                self.connectors[f"connector{block_i}"] = connector
                in_channels = connector.output_channels

            self.blocks_connected[f"block{block_i}"] = ConvBlock(
                in_channels,
                out_channels,
                norm_layer=norm_layer if 0 <= batchnorm_from <= block_i else None,
                padding=int(block_i < depth - 1),
            )
            self.blocks_channels += [out_channels]

    def forward(self, x, backbone_features):
        backbone_features = [] if backbone_features is None else backbone_features[::-1]

        outputs = [self.block0(x)]
        outputs += [self.block1(outputs[-1])]

        for block_i in range(2, self.depth):
            block = self.blocks_connected[f"block{block_i}"]
            output = outputs[-1]
            connector_name = f"connector{block_i}"
            if connector_name in self.connectors:
                stage_features = backbone_features.pop()
                connector = self.connectors[connector_name]
                output = connector(output, stage_features)
            outputs += [block(output)]

        return outputs[::-1]


class DeconvDecoder(nn.Module):
    def __init__(
        self,
        depth,
        encoder_blocks_channels,
        norm_layer,
        attend_from=-1,
        image_fusion=False,
    ):
        super(DeconvDecoder, self).__init__()
        self.image_fusion = image_fusion
        self.deconv_blocks = nn.ModuleList()

        in_channels = encoder_blocks_channels.pop()
        out_channels = in_channels
        for d in range(depth):
            out_channels = (
                encoder_blocks_channels.pop()
                if len(encoder_blocks_channels)
                else in_channels // 2
            )
            self.deconv_blocks.append(
                SEDeconvBlock(
                    in_channels,
                    out_channels,
                    norm_layer=norm_layer,
                    padding=0 if d == 0 else 1,
                    with_se=0 <= attend_from <= d,
                )
            )
            in_channels = out_channels

        if self.image_fusion:
            self.conv_attention = nn.Conv2d(out_channels, 1, kernel_size=1)
        self.to_rgb = nn.Conv2d(out_channels, 3, kernel_size=1)

    def forward(self, encoder_outputs, image, mask=None):
        output = encoder_outputs[0]
        for block, skip_output in zip(self.deconv_blocks[:-1], encoder_outputs[1:]):
            output = block(output, mask)
            output = output + skip_output
        output = self.deconv_blocks[-1](output, mask)

        if self.image_fusion:
            attention_map = torch.sigmoid(3.0 * self.conv_attention(output))
            output = attention_map * image + (1.0 - attention_map) * self.to_rgb(output)
        else:
            output = self.to_rgb(output)

        return output


class SEDeconvBlock(nn.Module):
    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size=4,
        stride=2,
        padding=1,
        norm_layer=nn.BatchNorm2d,
        activation=nn.ELU,
        with_se=False,
    ):
        super(SEDeconvBlock, self).__init__()
        self.with_se = with_se
        self.block = nn.Sequential(
            nn.ConvTranspose2d(
                in_channels, out_channels, kernel_size, stride=stride, padding=padding
            ),
            norm_layer(out_channels) if norm_layer is not None else nn.Identity(),
            activation(),
        )
        if self.with_se:
            self.se = MaskedChannelAttention(out_channels)

    def forward(self, x, mask=None):
        out = self.block(x)
        if self.with_se:
            out = self.se(out, mask)
        return out



from typing import Dict
from typing import List
from typing import Callable
from typing import NamedTuple
from cftool.misc import print_warning


class BackboneInfo(NamedTuple):
    fn: Callable
    out_channels: List[int]
    return_nodes: Dict[str, str]


backbone_info_dict: Dict[str, BackboneInfo] = {}


def register_backbone(
    name: str,
    out_channels: List[int],
    return_nodes: Dict[str, str],
) -> Callable:
    def _register(fn: Callable) -> Callable:
        registered = backbone_info_dict.get(name)
        if registered is not None:
            print_warning(
                f"'{name}' has already been registered "
                f"in the given global dict ({backbone_info_dict})"
            )
            return fn
        backbone_info_dict[name] = BackboneInfo(fn, out_channels, return_nodes)
        return fn

    return _register


__all__ = [
    "register_backbone",
    "backbone_info_dict",
]



import torch

import torch.nn as nn
import torch.nn.functional as F

from abc import ABCMeta
from torch import Tensor
from typing import Any
from typing import Dict
from typing import Type
from typing import Optional
from functools import partial
from torch.nn import Module
from cftool.misc import WithRegister

from .common import Lambda
from .hijacks import HijackLinear


activations: Dict[str, Type["Activation"]] = {}


class Activation(WithRegister["Activation"], Module, metaclass=ABCMeta):
    d = activations

    def __init__(self, **kwargs: Any):
        super().__init__()

    @classmethod
    def make(
        cls,
        name: Optional[str],
        config: Optional[Dict[str, Any]] = None,
    ) -> Module:
        if name is None:
            return nn.Identity()
        if config is None:
            config = {}
        if name.startswith("leaky_relu"):
            splits = name.split("_")
            if len(splits) == 3:
                config["negative_slope"] = float(splits[-1])
            config.setdefault("inplace", True)
            return nn.LeakyReLU(**config)
        if name.lower() == "relu":
            name = "ReLU"
            config.setdefault("inplace", True)
        base = cls.d.get(name, getattr(nn, name, None))
        if base is not None:
            return base(**config)
        func = getattr(torch, name, getattr(F, name, None))
        if func is None:
            raise NotImplementedError(
                "neither pytorch nor custom Activation "
                f"implemented activation '{name}'"
            )
        return Lambda(partial(func, **config), name)


@Activation.register("glu")
class GLU(Activation):
    def __init__(self, *, in_dim: int, bias: bool = True):
        super().__init__()
        self.linear = HijackLinear(in_dim, 2 * in_dim, bias)

    def forward(self, net: Tensor) -> Tensor:
        projection, gate = self.linear(net).chunk(2, dim=1)
        return projection * torch.sigmoid(gate)


@Activation.register("mish")
class Mish(Activation):
    def forward(self, net: Tensor) -> Tensor:
        return net * (torch.tanh(F.softplus(net)))


@Activation.register("atanh")
class ATanh(Activation):
    def __init__(self, *, eps: float = 1.0e-6):
        super().__init__()
        self.eps = eps

    def forward(self, net: Tensor) -> Tensor:
        return torch.atanh(torch.clamp(net, -1.0 + self.eps, 1.0 - self.eps))


@Activation.register("isoftplus")
class InverseSoftplus(Activation):
    def __init__(self, *, eps: float = 1.0e-6):
        super().__init__()
        self.eps = eps

    def forward(self, net: Tensor) -> Tensor:
        return torch.log(net.clamp_min(self.eps).exp() - 1.0)


@Activation.register("sign")
class Sign(Activation):
    def __init__(
        self,
        *,
        eps: float = 1.0e-12,
        randomize_at_zero: bool = False,
        differentiable: bool = True,
    ):
        super().__init__()
        self.eps = eps
        self.differentiable = differentiable
        self.randomize_at_zero = randomize_at_zero

    def forward(self, net: Tensor) -> Tensor:
        if self.randomize_at_zero:
            net = net + (2 * torch.empty_like(net).uniform_() - 1.0) * self.eps
        sign = torch.sign(net)
        if not self.differentiable:
            return sign
        return net + (sign - net).detach()


@Activation.register("one_hot")
class OneHot(Activation):
    def __init__(self, *, differentiable: bool = True):
        super().__init__()
        self.differentiable = differentiable

    def forward(self, net: Tensor) -> Tensor:
        maxed = torch.max(net, dim=1, keepdim=True)[0]
        one_hot = net * (net == maxed).to(torch.float32)
        if not self.differentiable:
            return one_hot
        return net + (one_hot - net).detach()


@Activation.register("sine")
class Sine(Activation):
    def __init__(self, *, w: float = 1.0):
        super().__init__()
        self.w = w

    def forward(self, net: Tensor) -> Tensor:
        return torch.sin(self.w * net)


@Activation.register("h_swish")
class HSwish(Activation):
    def __init__(self, *, inplace: bool = True):
        super().__init__()
        self.relu = nn.ReLU6(inplace=inplace)

    def forward(self, net: Tensor) -> Tensor:
        return net * (self.relu(net + 3.0) / 6.0)


@Activation.register("quick_gelu")
class QuickGELU(Activation):
    def forward(self, net: Tensor) -> Tensor:
        return net * torch.sigmoid(1.702 * net)


@Activation.register("geglu")
class GEGLU(Activation):
    def __init__(self, in_dim: int, out_dim: int):
        super().__init__()
        self.net = HijackLinear(in_dim, out_dim * 2)

    def forward(self, net: Tensor) -> Tensor:
        net, gate = self.net(net).chunk(2, dim=-1)
        return net * F.gelu(gate)


@Activation.register("diff_relu")
class DiffReLU(Activation):
    def forward(self, net: Tensor) -> Tensor:
        return net + (torch.relu(net) - net).detach()


__all__ = [
    "Activation",
]



import torch

import numpy as np

from abc import abstractmethod
from abc import ABC
from typing import Any
from typing import Dict
from typing import List
from typing import Tuple
from typing import Union
from typing import Optional
from cftool.misc import shallow_copy_dict
from cftool.array import to_device
from cftool.types import arr_type
from cftool.types import np_dict_type
from cftool.types import tensor_dict_type

from ..schema import IDataset
from ..schema import IDLModel
from ..schema import DataArgs
from ..schema import DataBundle
from ..schema import DataConfig
from ..schema import IDataLoader
from ..schema import DataProcessor
from ..constants import BATCH_INDICES_KEY
from ..misc.toolkit import eval_context
from ..misc.toolkit import get_device
from ..misc.toolkit import np_batch_to_tensor
from ..misc.toolkit import tensor_batch_to_np


TSplitSW = Tuple[Optional[np.ndarray], Optional[np.ndarray]]
TArrayDict = Union[np_dict_type, tensor_dict_type]
TArrayDataset = Union["IArrayDataset", "IArrayDictDataset"]


def get_weighted_indices(
    n: int,
    weights: Optional[np.ndarray],
    ensure_all_occur: bool = False,
) -> np.ndarray:
    indices = np.arange(n)
    if weights is not None:
        numbers = np.random.multinomial(n, weights)
        if ensure_all_occur:
            numbers += 1
        indices = indices.repeat(numbers)
    return indices


class IArrayDataset(IDataset):
    x: arr_type
    y: Optional[arr_type]
    processor: DataProcessor
    others: Optional[TArrayDict] = None


class IArrayDictDataset(IDataset):
    x: TArrayDict
    y: Optional[arr_type]
    processor: DataProcessor
    x_keys: List[str]


class ArrayLoader(IDataLoader):
    dataset: TArrayDataset

    cursor: int
    indices: np.ndarray

    shuffle: bool
    shuffle_backup: bool
    sample_weights: Optional[np.ndarray]

    def __init__(
        self,
        dataset: TArrayDataset,
        batch_size: int = 128,
        *,
        shuffle: bool = False,
        sample_weights: Optional[np.ndarray] = None,
    ):
        if sample_weights is not None and len(dataset) != len(sample_weights):
            raise ValueError(
                f"the number of data samples ({len(dataset)}) is not identical with "
                f"the number of sample weights ({len(sample_weights)})"
            )
        super().__init__(sample_weights=sample_weights)
        self.dataset = dataset
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.shuffle_backup = shuffle
        self.sample_weights = sample_weights

    def __iter__(self) -> "ArrayLoader":
        self.cursor = 0
        self.indices = get_weighted_indices(len(self.dataset), self.sample_weights)
        if self.shuffle:
            np.random.shuffle(self.indices)
        return self

    def __next__(self) -> TArrayDict:
        start = self.cursor
        if start >= len(self.dataset):
            raise StopIteration
        self.cursor += self.batch_size
        indices = self.indices[start : self.cursor]
        batch = self.dataset[indices]
        batch.setdefault(BATCH_INDICES_KEY, indices)
        return batch

    def disable_shuffle(self) -> None:
        self.shuffle = False

    def recover_shuffle(self) -> None:
        self.shuffle = self.shuffle_backup

    def copy(self) -> "ArrayLoader":
        return self.__class__(
            self.dataset,
            self.batch_size,
            shuffle=self.shuffle,
            sample_weights=self.sample_weights,
        )


class IArrayDataMixin(ABC):
    config: DataConfig
    bundle: DataBundle
    processor: DataProcessor
    train_dataset: TArrayDataset
    valid_dataset: Optional[TArrayDataset]
    train_weights: Optional[np.ndarray]
    valid_weights: Optional[np.ndarray]

    @property
    def train_kw(self) -> Dict[str, Any]:
        return dict(
            batch_size=self.config.batch_size,
            shuffle=self.config.shuffle_train,
            sample_weights=self.train_weights,
        )

    @property
    def valid_kw(self) -> Dict[str, Any]:
        return dict(
            batch_size=self.config.valid_batch_size or self.config.batch_size,
            shuffle=self.config.shuffle_valid,
            sample_weights=self.valid_weights,
        )

    def fit(
        self: "IArrayDataMixin",
        x_train: arr_type,
        y_train: Optional[arr_type] = None,
        x_valid: Optional[arr_type] = None,
        y_valid: Optional[arr_type] = None,
        train_others: Optional[TArrayDict] = None,
        valid_others: Optional[TArrayDict] = None,
        *args: Any,
        **kwargs: Any,
    ) -> "IArrayDataMixin":
        if train_others is not None:
            train_others = tensor_batch_to_np(train_others)
        if valid_others is not None:
            valid_others = tensor_batch_to_np(valid_others)
        return super().fit(  # type: ignore
            x_train,
            y_train,
            x_valid,
            y_valid,
            train_others,
            valid_others,
            *args,
            **kwargs,
        )

    def get_loaders(self) -> Tuple[ArrayLoader, Optional[ArrayLoader]]:
        if not self.processor.is_ready:
            raise ValueError(
                "`processor` should be ready before calling `initialize`, "
                "did you forget to call the `prepare` method first?"
            )
        if self.bundle is None:
            raise ValueError(
                "`bundle` property is not initialized, "
                "did you forget to call the `fit` method first?"
            )
        self.train_dataset = self.get_dataset(self.bundle.train_args)
        train_loader = ArrayLoader(self.train_dataset, **self.train_kw)
        if self.bundle.x_valid is None:
            valid_loader = None
        else:
            self.valid_dataset = self.get_dataset(self.bundle.valid_args)
            valid_loader = ArrayLoader(self.valid_dataset, **self.valid_kw)
        return train_loader, valid_loader

    @abstractmethod
    def get_dataset(self, data_args: DataArgs) -> TArrayDataset:
        pass


def predict_array_data(
    m: IDLModel,
    data: IArrayDataMixin,
    *,
    batch_size: Optional[int] = None,
    **predict_kwargs: Any,
) -> Any:
    if batch_size is not None:
        data.config.batch_size = batch_size
    loader = data.get_loaders()[0]
    results = []
    with eval_context(m):
        tensor_batcher = TensorBatcher(loader, get_device(m))
        for i, batch in enumerate(tensor_batcher):
            batch = shallow_copy_dict(batch)
            results.append(m.run(i, batch, **predict_kwargs))
    final = {}
    for k in results[0]:
        final[k] = torch.cat([rs[k] for rs in results], dim=0)
    return final


class TensorBatcher:
    def __init__(self, loader: IDataLoader, device: torch.device) -> None:
        self.loader = loader
        self.device = device

    def __len__(self) -> int:
        return len(self.loader)

    def __iter__(self) -> "TensorBatcher":
        self.loader.__iter__()
        return self

    def __next__(self) -> tensor_dict_type:
        npd = self.loader.__next__()
        batch = np_batch_to_tensor(npd)
        return to_device(batch, self.device)

    def to(self, device: torch.device) -> None:
        self.device = device

    def get_full_batch(self) -> tensor_dict_type:
        return np_batch_to_tensor(self.loader.get_full_batch())


__all__ = [
    "get_weighted_indices",
    "IArrayDataset",
    "IArrayDictDataset",
    "ArrayLoader",
    "IArrayDataMixin",
    "predict_array_data",
    "TensorBatcher",
]



import os

from abc import abstractmethod
from abc import ABCMeta
from typing import OrderedDict as OrderedDictType
from collections import OrderedDict

from ..core import Block
from ...schema import DLConfig


class InjectDefaultsMixin:
    _defaults: OrderedDictType

    def __init__(self) -> None:
        self._defaults = OrderedDict()

    def process_defaults(self, _defaults: OrderedDictType) -> None:
        for k, v in self._defaults.items():
            _defaults[k] = v


class TryLoadBlock(Block, metaclass=ABCMeta):
    # abstract

    @abstractmethod
    def try_load(self, folder: str) -> bool:
        pass

    @abstractmethod
    def from_scratch(self, config: DLConfig) -> None:
        pass

    @abstractmethod
    def dump_to(self, folder: str) -> None:
        pass

    # inheritance

    def build(self, config: DLConfig) -> None:
        if self.serialize_folder is not None:
            serialize_folder = os.path.join(self.serialize_folder, self.__identifier__)
            if self.try_load(serialize_folder):
                return
        self.from_scratch(config)

    def save_extra(self, folder: str) -> None:
        os.makedirs(folder, exist_ok=True)
        self.dump_to(folder)


__all__ = [
    "InjectDefaultsMixin",
    "TryLoadBlock",
]



import torch

from torch import Tensor


def append_dims(net: Tensor, ndim: int) -> Tensor:
    diff = ndim - net.ndim
    return net[(...,) + (None,) * diff]


def append_zero(net: Tensor) -> Tensor:
    return torch.cat([net, net.new_zeros([1])])


# credit : https://github.com/LuChengTHU/dpm-solver/blob/88a41b53b2d5d7c5a665bcb55efda5f5a6061223/dpm_solver_pytorch.py#L1105
def interpolate_fn(x: Tensor, xp: Tensor, yp: Tensor) -> Tensor:
    """
    A piecewise linear function y = f(x), using xp and yp as keypoints.
    We implement f(x) in a differentiable way (i.e. applicable for autograd).
    The function f(x) is well-defined for all x-axis. (For x beyond the bounds of xp, we use the outmost points of xp to define the linear function.)

    Args:
        x: PyTorch tensor with shape [N, C], where N is the batch size, C is the number of channels (we use C = 1 for DPM-Solver).
        xp: PyTorch tensor with shape [C, K], where K is the number of keypoints.
        yp: PyTorch tensor with shape [C, K].
    Returns:
        The function values f(x), with shape [N, C].
    """
    N, K = x.shape[0], xp.shape[1]
    all_x = torch.cat([x.unsqueeze(2), xp.unsqueeze(0).repeat((N, 1, 1))], dim=2)
    sorted_all_x, x_indices = torch.sort(all_x, dim=2)
    x_idx = torch.argmin(x_indices, dim=2)
    cand_start_idx = x_idx - 1
    start_idx = torch.where(
        torch.eq(x_idx, 0),
        torch.tensor(1, device=x.device),
        torch.where(
            torch.eq(x_idx, K),
            torch.tensor(K - 2, device=x.device),
            cand_start_idx,
        ),
    )
    end_idx = torch.where(
        torch.eq(start_idx, cand_start_idx), start_idx + 2, start_idx + 1
    )
    start_x = torch.gather(sorted_all_x, dim=2, index=start_idx.unsqueeze(2)).squeeze(2)
    end_x = torch.gather(sorted_all_x, dim=2, index=end_idx.unsqueeze(2)).squeeze(2)
    start_idx2 = torch.where(
        torch.eq(x_idx, 0),
        torch.tensor(0, device=x.device),
        torch.where(
            torch.eq(x_idx, K),
            torch.tensor(K - 2, device=x.device),
            cand_start_idx,
        ),
    )
    y_positions_expanded = yp.unsqueeze(0).expand(N, -1, -1)
    start_y = torch.gather(
        y_positions_expanded, dim=2, index=start_idx2.unsqueeze(2)
    ).squeeze(2)
    end_y = torch.gather(
        y_positions_expanded, dim=2, index=(start_idx2 + 1).unsqueeze(2)
    ).squeeze(2)
    cand = start_y + (x - start_x) * (end_y - start_y) / (end_x - start_x)
    return cand



import torch
import torch.nn as nn

from ..ops import MaskedChannelAttention
from ..modeling.unet import UNetDecoder
from ..modeling.unet import UNetEncoder


class ISEUNetV1(nn.Module):
    def __init__(
        self,
        depth,
        norm_layer=nn.BatchNorm2d,
        batchnorm_from=2,
        attend_from=3,
        image_fusion=False,
        ch=64,
        max_channels=512,
        backbone_from=-1,
        backbone_channels=None,
        backbone_mode="",
    ):
        super(ISEUNetV1, self).__init__()
        self.depth = depth
        self.encoder = UNetEncoder(
            depth,
            ch,
            norm_layer,
            batchnorm_from,
            max_channels,
            backbone_from,
            backbone_channels,
            backbone_mode,
        )
        self.decoder = UNetDecoder(
            depth,
            self.encoder.block_channels,
            norm_layer,
            attention_layer=MaskedChannelAttention,
            attend_from=attend_from,
            image_fusion=image_fusion,
        )

    def forward(self, image, mask, backbone_features=None):
        x = torch.cat((image, mask), dim=1)
        intermediates = self.encoder(x, backbone_features)
        output = self.decoder(intermediates, image, mask)
        return {"images": output}



import torch

import numpy as np
import torch.nn as nn

from torch import Tensor
from typing import Tuple
from typing import Optional
from cftool.types import tensor_dict_type

from .common import IAutoEncoder
from .common import AutoEncoderInit
from .common import AutoEncoderLPIPSWithDiscriminator
from ...schemas import CustomTrainStepLoss
from ....schema import IDLModel
from ....constants import INPUT_KEY
from ....constants import PREDICTIONS_KEY


class GaussianDistribution:
    key = "distribution"

    def __init__(self, net: Tensor, deterministic: bool = False):
        self.net = net
        self.device = net.device
        self.deterministic = deterministic
        self.mean, log_var = torch.chunk(net, 2, dim=1)
        self.log_var = torch.clamp(log_var, -30.0, 20.0)
        if self.deterministic:
            zeros = torch.zeros_like(self.mean)
            self.var = self.std = zeros
        else:
            self.std = torch.exp(0.5 * self.log_var)
            self.var = torch.exp(self.log_var)

    def sample(self) -> Tensor:
        std = self.std * torch.randn_like(self.mean)
        return self.mean + std

    def kl(self, other: Optional["GaussianDistribution"] = None) -> Tensor:
        if self.deterministic:
            return torch.tensor([0.0], device=self.device)
        if other is None:
            return 0.5 * torch.sum(
                torch.pow(self.mean, 2) + self.var - 1.0 - self.log_var,
                dim=[1, 2, 3],
            )
        return 0.5 * torch.sum(
            torch.pow(self.mean - other.mean, 2) / other.var
            + self.var / other.var
            - 1.0
            - self.log_var
            + other.log_var,
            dim=[1, 2, 3],
        )

    def nll(self, sample: Tensor, dims: Tuple[int, ...] = (1, 2, 3)) -> Tensor:
        if self.deterministic:
            return torch.tensor([0.0], device=self.device)
        return 0.5 * torch.sum(
            np.log(2.0 * np.pi)
            + self.log_var
            + torch.pow(sample - self.mean, 2) / self.var,
            dim=dims,
        )

    def mode(self) -> Tensor:
        return self.mean


class AutoEncoderKL(AutoEncoderInit):
    enc_double_channels = True

    def encode(self, net: Tensor) -> GaussianDistribution:
        net = self.generator.encode({INPUT_KEY: net})
        net = self.to_embedding(net)
        return GaussianDistribution(net)

    def decode(
        self,
        z: Tensor,
        *,
        resize: bool = True,
        apply_tanh: Optional[bool] = None,
    ) -> Tensor:
        net = self.from_embedding(z)
        kw = dict(resize=resize, apply_tanh=apply_tanh)
        net = self.generator.decode({INPUT_KEY: net}, **kw)  # type: ignore
        return net

    def forward(
        self,
        net: Tensor,
        *,
        sample_posterior: bool = True,
        apply_tanh: Optional[bool] = None,
    ) -> dict:
        distribution = self.encode(net)
        z = distribution.sample() if sample_posterior else distribution.mode()
        net = self.decode(z, apply_tanh=apply_tanh)
        return {PREDICTIONS_KEY: net, GaussianDistribution.key: distribution}


class AutoEncoderKLLoss(AutoEncoderLPIPSWithDiscriminator):
    def __init__(
        self,
        *,
        # kl
        log_var_init: float = 0.0,
        # common
        kl_weight: float = 1.0,
        d_loss: str = "hinge",
        d_loss_start_step: int = 50001,
        d_num_layers: int = 4,
        d_in_channels: int = 3,
        d_start_channels: int = 64,
        d_factor: float = 1.0,
        d_weight: float = 1.0,
        perceptual_weight: float = 1.0,
    ):
        super().__init__(
            kl_weight=kl_weight,
            d_loss=d_loss,
            d_loss_start_step=d_loss_start_step,
            d_num_layers=d_num_layers,
            d_in_channels=d_in_channels,
            d_start_channels=d_start_channels,
            d_factor=d_factor,
            d_weight=d_weight,
            perceptual_weight=perceptual_weight,
        )
        self.log_var = nn.Parameter(torch.ones(size=()) * log_var_init)

    def get_generator_loss(
        self,
        batch: tensor_dict_type,
        forward_results: tensor_dict_type,
        *,
        step: Optional[int],
        last_layer: nn.Parameter,
        cond: Optional[Tensor] = None,
        nll_weights: Optional[float] = None,
    ) -> CustomTrainStepLoss:
        inputs = batch[INPUT_KEY].contiguous()
        reconstructions = forward_results[PREDICTIONS_KEY].contiguous()
        loss_items = {}
        # recon & nll & kl loss
        recon_loss = torch.abs(inputs - reconstructions)
        with torch.no_grad():
            loss_items["l1"] = recon_loss.mean().item()
        if self.perceptual_weight > 0:
            p_loss = self.perceptual_loss(inputs, reconstructions)
            p_loss = self.perceptual_weight * p_loss
            recon_loss = recon_loss + p_loss
            with torch.no_grad():
                loss_items["perceptual"] = p_loss.mean().item()
        with torch.no_grad():
            loss_items["recon"] = recon_loss.mean().item()
        nll_loss = recon_loss / torch.exp(self.log_var) + self.log_var
        weighted_nll_loss = nll_loss
        if nll_weights is not None:
            weighted_nll_loss = nll_weights * nll_loss
        weighted_nll_loss = torch.sum(weighted_nll_loss) / weighted_nll_loss.shape[0]
        nll_loss = torch.sum(nll_loss) / nll_loss.shape[0]
        kl_loss = forward_results[GaussianDistribution.key].kl()
        kl_loss = torch.sum(kl_loss) / kl_loss.shape[0]
        loss = weighted_nll_loss + self.kl_weight * kl_loss
        loss_items["kl"] = self.kl_weight * kl_loss.item()
        # check discriminator start
        if step is not None and step < self.d_loss_start_step:
            return CustomTrainStepLoss(loss, loss_items)
        g_loss = self.g_loss(nll_loss, last_layer, loss_items, reconstructions, cond)
        loss = loss + g_loss
        return CustomTrainStepLoss(loss, loss_items)


@IDLModel.register("ae_kl")
class AutoEncoderKLModel(AutoEncoderKL, IAutoEncoder):  # type: ignore
    def __init__(
        self,
        img_size: int,
        in_channels: int,
        out_channels: int,
        inner_channels: int,
        latent_channels: int,
        channel_multipliers: Tuple[int, ...],
        *,
        embedding_channels: int,
        num_res_blocks: int,
        attention_resolutions: Tuple[int, ...] = (),
        dropout: float = 0.0,
        resample_with_conv: bool = True,
        attention_type: str = "spatial",
        # loss configs
        use_loss: bool = True,
        kl_weight: float = 1.0,
        log_var_init: float = 0.0,
        d_loss: str = "hinge",
        d_loss_start_step: int = 50001,
        d_num_layers: int = 4,
        d_in_channels: int = 3,
        d_start_channels: int = 64,
        d_factor: float = 1.0,
        d_weight: float = 1.0,
        perceptual_weight: float = 1.0,
        # train step configs
        grad_accumulate: int = 1,
    ):
        super().__init__(
            img_size,
            in_channels,
            out_channels,
            inner_channels,
            latent_channels,
            channel_multipliers,
            embedding_channels=embedding_channels,
            num_res_blocks=num_res_blocks,
            attention_resolutions=attention_resolutions,
            dropout=dropout,
            resample_with_conv=resample_with_conv,
            attention_type=attention_type,
        )
        if not use_loss:
            self.loss = None
        else:
            self.loss = AutoEncoderKLLoss(
                kl_weight=kl_weight,
                log_var_init=log_var_init,
                d_loss=d_loss,
                d_loss_start_step=d_loss_start_step,
                d_num_layers=d_num_layers,
                d_in_channels=d_in_channels,
                d_start_channels=d_start_channels,
                d_factor=d_factor,
                d_weight=d_weight,
                perceptual_weight=perceptual_weight,
            )
        self.setup(img_size, grad_accumulate, embedding_channels, channel_multipliers)


__all__ = [
    "AutoEncoderKL",
]



import math

from typing import Optional
from cftool.misc import Incrementer

from .schema import TrainerMonitor


@TrainerMonitor.register("basic")
class BasicMonitor(TrainerMonitor):
    def __init__(self, patience: int = 25):  # type: ignore
        super().__init__()
        self.patience = patience
        self.num_snapshot = 0
        self.best_score = -math.inf
        self.worst_score: Optional[float] = None

    def snapshot(self, new_score: float) -> bool:
        self.num_snapshot += 1
        if self.worst_score is None:
            self.worst_score = new_score
        else:
            self.worst_score = min(new_score, self.worst_score)
        if new_score > self.best_score:
            self.best_score = new_score
            return True
        return False

    def check_terminate(self, new_score: float) -> bool:
        if self.num_snapshot <= self.patience:
            return False
        if self.worst_score is None:
            return False
        return new_score <= self.worst_score

    def punish_extension(self) -> None:
        return None


@TrainerMonitor.register("mean_std")
class MeanStdMonitor(BasicMonitor):
    def __init__(
        self,
        *,
        patience: int = 5,
        window_size: int = 25,
        overfit_tolerance: float = 25.0,
    ):
        super().__init__()
        self.patience = patience
        self.overfit_tolerance = overfit_tolerance
        self.best_score = -math.inf
        self.overfit_level = 0.0
        self._incrementer = Incrementer(window_size)

    def snapshot(self, new_score: float) -> bool:
        self._incrementer.update(new_score)
        mean, std = self._incrementer.mean, self._incrementer.std
        std = max(std, 1.0e-8)
        if new_score < mean - std:
            max_decrease = self.overfit_tolerance / self.patience
            decrease = min(max_decrease, (mean - new_score) / std + 1.0)
            self.overfit_level += decrease
        elif new_score > mean + std:
            improvement = (new_score - mean) / std - 1.0
            self.overfit_level = max(0.0, self.overfit_level - improvement)
        return super().snapshot(new_score)

    def check_terminate(self, new_score: float) -> bool:
        if self.num_snapshot <= 10:
            return False
        if self.overfit_level >= self.overfit_tolerance:
            return True
        return False


@TrainerMonitor.register("plateau")
class PlateauMonitor(TrainerMonitor):
    def __init__(
        self,
        *,
        patience: float = 5.0,
        window_size: int = 25,
        plateau_tolerance: float = 25.0,
        plateau_threshold: float = 0.2,
    ):
        super().__init__()
        self.patience = patience
        self.window_size = window_size
        self.plateau_tolerance = plateau_tolerance
        self.plateau_threshold = plateau_threshold
        self.num_snapshot = 0
        self.plateau_level = 0.0
        self._incrementer = Incrementer(window_size)

    @property
    def max_plateau_increase(self) -> float:
        return self.plateau_tolerance / self.patience

    def snapshot(self, new_score: float) -> bool:
        self.num_snapshot += 1
        self._incrementer.update(new_score)
        if self.num_snapshot <= self.window_size:
            return False
        mean, std = self._incrementer.mean, self._incrementer.std
        ratio = max(abs(new_score - mean) / max(std, 1.0e-8), 1.0e-8)
        if ratio < self.plateau_threshold:
            plateau = min(
                self.max_plateau_increase,
                1.0 / ratio - 1.0 / self.plateau_threshold,
            )
            self.plateau_level += plateau
        return False

    def check_terminate(self, new_score: float) -> bool:
        if self.plateau_level >= self.plateau_tolerance:
            return True
        return False

    def punish_extension(self) -> None:
        self.plateau_level += self.max_plateau_increase / 5.0


@TrainerMonitor.register("conservative")
class ConservativeMonitor(TrainerMonitor):
    def snapshot(self, new_score: float) -> bool:
        return True

    def check_terminate(self, new_score: float) -> bool:
        return False

    def punish_extension(self) -> None:
        pass


@TrainerMonitor.register("lazy")
class LazyMonitor(TrainerMonitor):
    def snapshot(self, new_score: float) -> bool:
        return False

    def check_terminate(self, new_score: float) -> bool:
        return False

    def punish_extension(self) -> None:
        pass


__all__ = [
    "BasicMonitor",
    "MeanStdMonitor",
    "PlateauMonitor",
    "ConservativeMonitor",
    "LazyMonitor",
]



import torch.nn as nn
import torch.nn.init as init

from torch import no_grad
from torch import Tensor
from typing import Dict
from typing import List
from typing import Optional

from .fcnn import FCNN
from typing import Dict
from typing import Optional

from .base import MLModel
from ...schema import MLEncoderSettings
from ...schema import MLGlobalEncoderSettings


rnn_dict = {
    "LSTM": nn.LSTM,
    "GRU": nn.GRU,
    "RNN": nn.RNN,
}


@MLModel.register("rnn")
class RNN(MLModel):
    def __init__(
        self,
        input_dim: int,
        output_dim: int,
        num_history: int = 1,
        cell: str = "GRU",
        num_layers: int = 1,
        hidden_size: int = 256,
        bidirectional: bool = False,
        hidden_units: Optional[List[int]] = None,
        *,
        mapping_type: str = "basic",
        bias: bool = True,
        activation: str = "ReLU",
        batch_norm: bool = False,
        dropout: float = 0.0,
        encoder_settings: Optional[Dict[str, MLEncoderSettings]] = None,
        global_encoder_settings: Optional[MLGlobalEncoderSettings] = None,
    ):
        super().__init__(
            encoder_settings=encoder_settings,
            global_encoder_settings=global_encoder_settings,
        )
        if self.encoder is not None:
            input_dim += self.encoder.dim_increment
        self.input_dim = input_dim
        self.output_dim = output_dim
        self.num_history = num_history
        rnn_dim = self._init_rnn(cell, num_layers, hidden_size, bidirectional)
        self.head = FCNN(
            rnn_dim,
            output_dim,
            1,
            hidden_units,
            mapping_type=mapping_type,
            bias=bias,
            activation=activation,
            batch_norm=batch_norm,
            dropout=dropout,
        )

    def _init_rnn(
        self,
        cell: str,
        num_layers: int,
        hidden_size: int,
        bidirectional: bool,
    ) -> int:
        rnn_base = rnn_dict[cell]
        input_dimensions = [self.input_dim]
        cell_config = {
            "batch_first": True,
            "hidden_size": hidden_size,
            "bidirectional": bidirectional,
        }
        rnn_dim = hidden_size * (1 + int(bidirectional))
        input_dimensions += [rnn_dim] * (num_layers - 1)
        rnn_list = []
        for dim in input_dimensions:
            rnn = rnn_base(dim, **cell_config)
            with no_grad():
                for name, param in rnn.named_parameters():
                    if "weight" in name:
                        init.orthogonal_(param)
                    elif "bias" in name:
                        init.zeros_(param)
            rnn_list.append(rnn)
        self.rnn_list = nn.ModuleList(rnn_list)
        return rnn_dim

    def forward(self, net: Tensor) -> Tensor:
        for rnn in self.rnn_list:
            net, final_state = rnn(net, None)
        return self.head(net[:, -1])


__all__ = [
    "RNN",
]



import json

import numpy as np

from enum import Enum
from torch import Tensor
from typing import Any
from typing import Dict
from typing import Tuple
from typing import Union
from typing import Optional
from cftool.misc import print_info
from cftool.array import tensor_dict_type

from .ddpm import make_condition_model
from .ddpm import DDPM
from .utils import CROSS_ATTN_TYPE
from ..ae.kl import GaussianDistribution
from ....zoo import DLZoo
from ....schema import IDLModel
from ....misc.toolkit import freeze
from ....misc.toolkit import get_tensors
from ....misc.toolkit import download_static
from ....modules.blocks import IHook
from ....modules.blocks import LoRAManager


@IDLModel.register("ldm")
class LDM(DDPM):
    def __init__(
        self,
        img_size: int,
        # unet
        in_channels: int,
        out_channels: int,
        *,
        num_heads: Optional[int] = 8,
        num_head_channels: Optional[int] = None,
        use_spatial_transformer: bool = False,
        num_transformer_layers: int = 1,
        context_dim: Optional[int] = None,
        signal_dim: int = 2,
        start_channels: int = 320,
        num_res_blocks: int = 2,
        attention_downsample_rates: Tuple[int, ...] = (1, 2, 4),
        dropout: float = 0.0,
        channel_multipliers: Tuple[int, ...] = (1, 2, 4, 4),
        resample_with_conv: bool = True,
        resample_with_resblock: bool = False,
        use_scale_shift_norm: bool = False,
        num_classes: Optional[int] = None,
        use_linear_in_transformer: bool = False,
        use_checkpoint: bool = False,
        attn_split_chunk: Optional[int] = None,
        tome_info: Optional[Dict[str, Any]] = None,
        # first stage
        first_stage: str,
        first_stage_config: Optional[Dict[str, Any]] = None,
        first_stage_scale_factor: float = 1.0,
        # diffusion
        ema_decay: Optional[float] = None,
        use_num_updates_in_ema: bool = True,
        parameterization: str = "eps",
        ## condition
        condition_type: str = CROSS_ATTN_TYPE,
        condition_model: Optional[str] = None,
        use_first_stage_as_condition: bool = False,
        condition_config: Optional[Dict[str, Any]] = None,
        condition_learnable: bool = False,
        use_pretrained_condition: bool = False,
        ## noise schedule
        v_posterior: float = 0.0,
        timesteps: int = 1000,
        given_betas: Optional[np.ndarray] = None,
        beta_schedule: str = "linear",
        linear_start: float = 1.0e-4,
        linear_end: float = 2.0e-2,
        cosine_s: float = 8.0e-3,
        ## loss
        loss_type: str = "l2",
        l_simple_weight: float = 1.0,
        original_elbo_weight: float = 0.0,
        learn_log_var: bool = False,
        log_var_init: float = 0.0,
        ## sampling
        sampler: str = "ddim",
        sampler_config: Optional[Dict[str, Any]] = None,
    ):
        self.use_first_stage_as_condition = use_first_stage_as_condition
        if use_first_stage_as_condition:
            if condition_learnable:
                raise ValueError(
                    "should not use ae as condition model "
                    "when `condition_learnable` is set to True"
                )
            condition_model = None
        super().__init__(
            img_size,
            in_channels,
            out_channels,
            num_heads=num_heads,
            num_head_channels=num_head_channels,
            use_spatial_transformer=use_spatial_transformer,
            num_transformer_layers=num_transformer_layers,
            context_dim=context_dim,
            signal_dim=signal_dim,
            start_channels=start_channels,
            num_res_blocks=num_res_blocks,
            attention_downsample_rates=attention_downsample_rates,
            dropout=dropout,
            channel_multipliers=channel_multipliers,
            resample_with_conv=resample_with_conv,
            resample_with_resblock=resample_with_resblock,
            use_scale_shift_norm=use_scale_shift_norm,
            num_classes=num_classes,
            use_linear_in_transformer=use_linear_in_transformer,
            use_checkpoint=use_checkpoint,
            attn_split_chunk=attn_split_chunk,
            tome_info=tome_info,
            ema_decay=ema_decay,
            use_num_updates_in_ema=use_num_updates_in_ema,
            parameterization=parameterization,
            condition_type=condition_type,
            condition_model=condition_model,
            condition_config=condition_config,
            condition_learnable=condition_learnable,
            use_pretrained_condition=use_pretrained_condition,
            v_posterior=v_posterior,
            timesteps=timesteps,
            given_betas=given_betas,
            beta_schedule=beta_schedule,
            linear_start=linear_start,
            linear_end=linear_end,
            cosine_s=cosine_s,
            loss_type=loss_type,
            l_simple_weight=l_simple_weight,
            original_elbo_weight=original_elbo_weight,
            learn_log_var=learn_log_var,
            log_var_init=log_var_init,
            sampler=sampler,
            sampler_config=sampler_config,
        )
        first_stage_kw = first_stage_config or {}
        m = DLZoo.load_pipeline(first_stage, **first_stage_kw)
        self.first_stage = freeze(m.build_model.model)
        self.scale_factor = first_stage_scale_factor
        # condition
        if use_first_stage_as_condition:
            # avoid recording duplicate parameters
            self.condition_model = [make_condition_model(first_stage, self.first_stage)]
        # sanity check
        embedding_channels = self.first_stage.embedding_channels
        if in_channels != embedding_channels and condition_type == CROSS_ATTN_TYPE:
            raise ValueError(
                f"`in_channels` ({in_channels}) should be identical with the "
                f"`embedding_channels` ({embedding_channels}) of the "
                f"first_stage model ({first_stage})"
            )
        if out_channels != embedding_channels:
            raise ValueError(
                f"`out_channels` ({out_channels}) should be identical with the "
                f"`embedding_channels` ({embedding_channels}) of the "
                f"first_stage model ({first_stage})"
            )

    def decode(
        self,
        z: Tensor,
        *,
        cond: Optional[Any] = None,
        num_steps: Optional[int] = None,
        start_step: Optional[int] = None,
        verbose: bool = True,
        **kwargs: Any,
    ) -> Tensor:
        latent = super().decode(
            z,
            cond=cond,
            num_steps=num_steps,
            start_step=start_step,
            verbose=verbose,
            **kwargs,
        )
        net = self._from_latent(latent)
        return net

    def _preprocess(self, net: Tensor, *, deterministic: bool = False) -> Tensor:
        net = self.first_stage.encode(net)
        if isinstance(net, GaussianDistribution):
            net = net.mode() if deterministic else net.sample()
        net = self.scale_factor * net
        return net

    def _from_latent(self, latent: Tensor) -> Tensor:
        latent = latent / self.scale_factor
        return self.first_stage.decode(latent, resize=False)

    def _get_cond(self, cond: Any) -> Tensor:
        if isinstance(self.condition_model, list):
            return self.condition_model[0](cond)
        return super()._get_cond(cond)


class SDLoRAMode(str, Enum):
    UNET = "unet"
    UNET_EXTENDED = "unet_extended"


def convert_lora(inp: Union[str, tensor_dict_type]) -> tensor_dict_type:
    inp = get_tensors(inp)
    with open(download_static("sd_lora_mapping", extension="json"), "r") as f:
        mapping = json.load(f)
    return {mapping[k]: v for k, v in inp.items()}


@IDLModel.register("sd")
class StableDiffusion(LDM):
    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.lora_manager = LoRAManager()

    @property
    def has_lora(self) -> bool:
        return self.lora_manager.injected

    def load_lora(self, key: str, *, path: str) -> None:
        print_info(f"loading '{key}' from '{path}'")
        d = get_tensors(path)
        wk = "lora_unet_down_blocks_0_attentions_0_proj_in.lora_down.weight"
        rank = d[wk].shape[0]
        mode = SDLoRAMode.UNET if "res" not in d else SDLoRAMode.UNET_EXTENDED
        inject_text_encoder = "lora_te_text_model_encoder_layers_0_mlp_fc1.alpha" in d
        print_info(
            f"preparing lora (rank={rank}, mode={mode}, "
            f"inject_text_encoder={inject_text_encoder})"
        )
        self.prepare_lora(key, rank, mode=mode, inject_text_encoder=inject_text_encoder)
        print_info("loading weights")
        self.lora_manager.load_pack_with(key, convert_lora(d))
        print_info(f"finished loading '{key}'")

    def prepare_lora(
        self,
        key: str,
        rank: int,
        *,
        mode: SDLoRAMode = SDLoRAMode.UNET,
        inject_text_encoder: bool = True,
    ) -> None:
        target_ancestors = {"SpatialTransformer", "GEGLU"}
        if mode == SDLoRAMode.UNET_EXTENDED:
            target_ancestors.add("ResBlock")
        if inject_text_encoder:
            target_ancestors.add("TeTEncoder")
        self.lora_manager.prepare(
            self,
            key=key,
            rank=rank,
            target_ancestors=target_ancestors,
        )

    def inject_lora(self, *keys: str) -> None:
        self.lora_manager.inject(self, *keys)

    def cleanup_lora(self) -> None:
        self.lora_manager.cleanup(self)

    def set_lora_scales(self, scales: Dict[str, float]) -> None:
        self.lora_manager.set_scales(scales)

    def get_lora_checkpoints(self) -> Dict[str, Optional[IHook]]:
        return self.lora_manager.checkpoint(self)

    def restore_lora_from(self, hooks: Dict[str, Optional[IHook]]) -> None:
        return self.lora_manager.restore(self, hooks)


__all__ = [
    "LDM",
    "SDLoRAMode",
    "StableDiffusion",
    "convert_lora",
]



from typing import List
from typing import Type
from typing import Optional
from torchvision.datasets import MNIST

from .api import TorchData
from .api import TorchDataConfig
from ..blocks import HWCToCHWBlock
from ..blocks import ToNumpyBlock
from ..blocks import TupleToBatchBlock
from ..blocks import StaticNormalizeBlock
from ...schema import IDataBlock
from ...schema import DataProcessorConfig
from ...parameters import OPT


def mnist_data(
    config: Optional[TorchDataConfig] = None,
    processor_config: Optional[DataProcessorConfig] = None,
    *,
    cache_root: str = OPT.data_cache_dir,
    additional_blocks: Optional[List[IDataBlock]] = None,
) -> TorchData:
    if processor_config is None:
        processor_config = DataProcessorConfig()
    if processor_config.block_names is None:
        processor_config.set_blocks(
            TupleToBatchBlock(),
            ToNumpyBlock(),
            StaticNormalizeBlock(),
            HWCToCHWBlock(),
        )
    if additional_blocks is not None:
        processor_config.add_blocks(*additional_blocks)
    train_data = MNIST(cache_root, download=True)
    valid_data = MNIST(cache_root, train=False, download=True)
    return TorchData.build(
        train_data,
        valid_data,
        config=config,
        processor_config=processor_config,
    )


__all__ = [
    "mnist_data",
]



import math
import torch

import torch.nn as nn
import torch.nn.functional as F

from torch import Tensor
from typing import Tuple
from typing import Union
from typing import Callable
from typing import Optional
from cftool.types import tensor_dict_type

from ...constants import LABEL_KEY
from ...modules.blocks import Lambda
from ...modules.blocks import Activation
from ...modules.blocks import ChannelPadding


class SirenLayer(nn.Module):
    def __init__(
        self,
        in_dim: int,
        out_dim: int,
        *,
        w_sin: float = 1.0,
        c_init: float = 6.0,
        bias: bool = True,
        is_first: bool = False,
        activation: Optional[nn.Module] = None,
    ):
        super().__init__()
        self.in_dim = in_dim
        self.is_first = is_first
        # siren weight
        weight = torch.zeros(out_dim, in_dim)
        w_std = (1.0 / in_dim) if is_first else (math.sqrt(c_init / in_dim) / w_sin)
        weight.uniform_(-w_std, w_std)
        self.weight = nn.Parameter(weight)
        # siren bias
        self.bias = None
        if bias:
            bias_value = torch.zeros(out_dim).uniform_(-w_std, w_std)
            self.bias = nn.Parameter(bias_value)
        # siren activation
        if activation is not None:
            self.activation = activation
        else:
            self.activation = Activation.make("sine", {"w": w_sin})

    def forward(self, net: Tensor) -> Tensor:
        net = F.linear(net, self.weight, self.bias)
        net = self.activation(net)
        return net


class Modulator(nn.Module):
    def __init__(self, in_dim: int, latent_dim: int, num_layers: int):
        super().__init__()
        blocks = []
        for i in range(num_layers):
            dim = in_dim if i == 0 else (latent_dim + in_dim)
            blocks.append(
                nn.Sequential(
                    nn.Linear(dim, latent_dim),
                    nn.ReLU(inplace=True),
                )
            )
        self.blocks = nn.ModuleList(blocks)

    def forward(self, z: Tensor) -> Tuple[Tensor, ...]:
        net = z
        nets = []
        for block in self.blocks:
            net = block(net)
            nets.append(net)
            net = torch.cat((net, z), dim=1)
        return tuple(nets)


def make_grid(size: int, in_dim: int, device: Optional[torch.device] = None) -> Tensor:
    tensors = [torch.linspace(-1.0, 1.0, steps=size, device=device)] * in_dim
    grid = torch.stack(torch.meshgrid(*tensors, indexing="ij"), dim=-1)
    return grid.view(1, -1, grid.shape[-1])


class Siren(nn.Module):
    grid: Optional[Tensor]

    def __init__(
        self,
        size: Optional[int],
        in_dim: int,
        out_dim: int,
        latent_dim: int,
        *,
        num_layers: int = 4,
        w_sin: float = 1.0,
        w_sin_initial: float = 30.0,
        bias: bool = True,
        final_activation: Optional[str] = None,
        keep_edge: bool = True,
        use_modulator: bool = True,
    ):
        super().__init__()
        self.in_dim = in_dim
        self.num_layers = num_layers
        self.latent_dim = latent_dim
        # siren blocks
        blocks = []
        for i in range(num_layers):
            is_first = i == 0
            i_w_sin = w_sin_initial if is_first else w_sin
            i_in_dim = in_dim if is_first else latent_dim
            blocks.append(
                SirenLayer(
                    in_dim=i_in_dim,
                    out_dim=latent_dim,
                    w_sin=i_w_sin,
                    bias=bias,
                    is_first=is_first,
                )
            )
        self.blocks = nn.ModuleList(blocks)
        # siren grid
        self.keep_edge = keep_edge
        if size is None:
            self.grid = None
        else:
            grid = make_grid(size, in_dim)
            self.register_buffer("grid", grid)
        # latent modulator
        if not use_modulator:
            self.modulator = None
        else:
            self.modulator = Modulator(latent_dim, latent_dim, num_layers)
        # head
        self.head = SirenLayer(
            in_dim=latent_dim,
            out_dim=out_dim,
            w_sin=w_sin,
            bias=bias,
            activation=Activation.make(final_activation),
        )

    def forward(
        self,
        inp: Union[Tensor, Tuple[Tensor, ...]],
        *,
        init: Optional[Tensor] = None,
        size: Optional[int] = None,
    ) -> Tensor:
        # get mods
        if self.modulator is None:
            mods = inp
            device = inp[0].device
        else:
            mods = self.modulator(inp)
            device = inp.device  # type: ignore
        # get siren input
        if init is not None:
            net = init
        else:
            if self.grid is None and size is None:
                raise ValueError("`init` should be provided for None sized `Siren`")
            else:
                if size is None:
                    grid = self.grid
                else:
                    grid = make_grid(size, self.in_dim, device)
                    if not self.keep_edge:
                        grid = grid[1:-1]
                net = grid.repeat(len(inp), 1, 1)  # type: ignore
        # forward process
        for block, mod in zip(self.blocks, mods):
            net = block(net) * mod.unsqueeze(1)
        return self.head(net)


# image usages


def img_siren_head(size: int, out_channels: int) -> Callable[[Tensor], Tensor]:
    def _head(t: Tensor) -> Tensor:
        t = t.view(-1, size, size, out_channels)
        t = t.permute(0, 3, 1, 2)
        return t

    return _head


class ImgSiren(nn.Module):
    def __init__(
        self,
        img_size: int,
        out_channels: int,
        latent_dim: int = 256,
        num_classes: Optional[int] = None,
        conditional_dim: int = 16,
        *,
        num_layers: int = 4,
        w_sin: float = 1.0,
        w_sin_initial: float = 30.0,
        bias: bool = True,
        final_activation: Optional[str] = None,
    ):
        super().__init__()
        self.out_channels = out_channels
        # condition
        self.cond_padding = None
        if num_classes is not None:
            self.cond_padding = ChannelPadding(
                latent_dim,
                conditional_dim,
                is_1d=True,
                num_classes=num_classes,
            )
        # siren
        self.siren = Siren(
            img_size,
            2,
            out_channels,
            latent_dim,
            num_layers=num_layers,
            w_sin=w_sin,
            w_sin_initial=w_sin_initial,
            bias=bias,
            final_activation=final_activation,
        )
        # head
        self.head = Lambda(img_siren_head(img_size, out_channels), name="head")

    def forward(self, net: Tensor, batch: tensor_dict_type) -> Tensor:
        if self.cond_padding is not None:
            net = self.cond_padding(net, batch[LABEL_KEY].view(-1))
        net = self.siren(net)
        return self.head(net)

    def decode(
        self,
        z: Tensor,
        *,
        labels: Optional[Tensor],
        size: Optional[int] = None,
    ) -> Tensor:
        if self.cond_padding is not None:
            if labels is None:
                msg = "`labels` should be provided in conditional `ImgSiren`"
                raise ValueError(msg)
            z = self.cond_padding(z, labels)
        net = self.siren(z, size=size)
        if size is None:
            return self.head(net)
        return img_siren_head(size, self.out_channels)(net)


__all__ = [
    "Siren",
    "ImgSiren",
    "make_grid",
    "img_siren_head",
]



import os

from cftool.cv import save_images
from cftool.array import to_device

from .general import ImageCallback
from ..misc.toolkit import eval_context
from ..schema import ITrainer
from ..schema import TrainerCallback
from ..constants import INPUT_KEY
from ..constants import LABEL_KEY
from ..models.schemas.cv import GeneratorMixin


@TrainerCallback.register("ldm")
@TrainerCallback.register("ddpm")
@TrainerCallback.register("ae_kl")
@TrainerCallback.register("ae_vq")
@TrainerCallback.register("gan")
@TrainerCallback.register("vae")
@TrainerCallback.register("vae2d")
@TrainerCallback.register("style_vae")
@TrainerCallback.register("generator")
class GeneratorCallback(ImageCallback):
    def __init__(self, num_keep: int = 25, num_interpolations: int = 16):
        super().__init__(num_keep)
        self.num_interpolations = num_interpolations

    def log_artifacts(self, trainer: ITrainer) -> None:
        if not self.is_local_rank_0:
            return None
        batch = next(iter(trainer.validation_loader))
        batch = to_device(batch, trainer.device)
        original = batch[INPUT_KEY]
        model = trainer.model
        if not isinstance(model, GeneratorMixin):
            msg = "`GeneratorCallback` is only compatible with `GeneratorMixin`"
            raise ValueError(msg)
        is_conditional = getattr(model, "is_conditional", False)
        labels = None if not is_conditional else batch[LABEL_KEY]
        image_folder = self._prepare_folder(trainer)
        # original
        save_images(original, os.path.join(image_folder, "original.png"))
        # reconstruct
        if model.can_reconstruct:
            with eval_context(model):
                reconstructed = model.reconstruct(original, labels=labels)
            save_images(reconstructed, os.path.join(image_folder, "reconstructed.png"))
        # sample
        with eval_context(model):
            sampled = model.sample(len(original))
        save_images(sampled, os.path.join(image_folder, "sampled.png"))
        # interpolation
        with eval_context(model):
            interpolations = model.interpolate(self.num_interpolations)
        save_images(interpolations, os.path.join(image_folder, "interpolations.png"))
        # conditional sampling
        if getattr(model, "num_classes", None) is None:
            return None
        cond_folder = os.path.join(image_folder, "conditional")
        os.makedirs(cond_folder, exist_ok=True)
        with eval_context(model):
            for i in range(model.num_classes or 0):
                sampled = model.sample(len(original), class_idx=i)
                interpolations = model.interpolate(len(original), class_idx=i)
                save_images(sampled, os.path.join(cond_folder, f"sampled_{i}.png"))
                path = os.path.join(cond_folder, f"interpolations_{i}.png")
                save_images(interpolations, path)


@TrainerCallback.register("siren_gan")
@TrainerCallback.register("siren_vae")
@TrainerCallback.register("sized_generator")
class SizedGeneratorCallback(GeneratorCallback):
    def log_artifacts(self, trainer: ITrainer) -> None:
        if not self.is_local_rank_0:
            return None
        super().log_artifacts(trainer)
        model = trainer.model
        image_folder = self._prepare_folder(trainer, check_num_keep=False)
        sample_method = getattr(model, "sample", None)
        if sample_method is None:
            raise ValueError(
                "`sample` should be implemented when `SizedGeneratorCallback` is used "
                "(and the `sample` method should support accepting `size` kwarg)"
            )
        resolution = getattr(model, "img_size", 32)
        with eval_context(model):
            for i in range(1, 4):
                size = resolution * 2**i
                batch_size = 1 if size > 256 else 4
                sampled = sample_method(batch_size, size=size).cpu()
                path = os.path.join(image_folder, f"sampled_{size}x{size}.png")
                save_images(sampled, path)


__all__ = [
    "GeneratorCallback",
    "SizedGeneratorCallback",
]
